import os
from functools import partial
from typing import Callable
from unittest.mock import MagicMock, patch

import pytest

import dataquality
from dataquality import config
from dataquality.core.auth import GALILEO_AUTH_METHOD
from dataquality.exceptions import GalileoException
from tests.exceptions import LoginInvoked
from tests.utils.mock_request import (
    EXISTING_PROJECT,
    EXISTING_RUN,
    mocked_create_project_run,
    mocked_get_project_run,
    mocked_login,
    mocked_login_requests,
    mocked_missing_project_run,
    mocked_missing_run,
)


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    dataquality.init(task_type="text_classification")
    assert config.current_run_id
    assert config.current_project_id


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_reset_logger_config(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    dataquality.init(task_type="text_classification")
    assert config.current_run_id
    assert config.current_project_id
    dataquality.set_labels_for_run(["a", "b", "c", "d"])
    dataquality.init(task_type="text_classification")
    assert not dataquality.get_data_logger().logger_config.labels


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_private(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    dataquality.init(task_type="text_classification", is_public=False)
    assert config.current_run_id
    assert config.current_project_id
    mock_create_project_call = mock_requests_post.call_args_list[0]
    assert mock_create_project_call.assert_called_with(is_public=False)


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_existing_project(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Tests calling init passing in an existing project"""
    set_test_config(current_project_id=None, current_run_id=None)
    dataquality.init(task_type="text_classification", project_name=EXISTING_PROJECT)
    assert config.current_run_id
    assert config.current_project_id


@patch("requests.get", side_effect=mocked_missing_project_run)
@patch("requests.post", side_effect=mocked_create_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_new_project(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Tests calling init passing in a new project"""
    set_test_config(current_project_id=None, current_run_id=None)
    dataquality.init(task_type="text_classification", project_name="new_proj")
    assert config.current_run_id
    assert config.current_project_id


@patch("requests.get", side_effect=mocked_missing_run)
@patch("requests.post", side_effect=mocked_create_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_existing_project_new_run(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Tests calling init with an existing project but a new run"""
    set_test_config(current_project_id=None, current_run_id=None)
    dataquality.init(
        task_type="text_classification",
        project_name=EXISTING_PROJECT,
        run_name="new_run",
    )
    assert config.current_run_id
    assert config.current_project_id


@patch("requests.get", side_effect=mocked_get_project_run)
@patch("requests.post", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_existing_project_run(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Tests calling init with an existing project and existing run"""
    set_test_config(current_project_id=None, current_run_id=None)
    dataquality.init(
        task_type="text_classification",
        project_name=EXISTING_PROJECT,
        run_name=EXISTING_RUN,
    )
    assert config.current_run_id
    assert config.current_project_id


@patch("requests.get", side_effect=mocked_missing_project_run)
@patch("requests.post", side_effect=mocked_create_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_new_project_run(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    """Tests calling init with a new project and new run"""
    set_test_config(current_project_id=None, current_run_id=None)
    dataquality.init(
        task_type="text_classification", project_name="new_proj", run_name="new_run"
    )
    assert config.current_run_id
    assert config.current_project_id


@patch("requests.get", side_effect=mocked_missing_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_only_run(
    mock_valid_user: MagicMock, mock_requests_get: MagicMock, set_test_config: Callable
) -> None:
    """Tests calling init only passing in a run"""
    set_test_config(current_project_id=None, current_run_id=None)
    dataquality.init(task_type="text_classification", run_name="a_run")
    assert not config.current_run_id
    assert not config.current_project_id


@patch("dataquality.core.init.login", side_effect=LoginInvoked)
def test_init_no_token_login(mock_login: MagicMock, set_test_config: Callable) -> None:
    set_test_config(token=None)
    with pytest.raises(LoginInvoked):
        # When no token is passed in we should call login
        dataquality.init(task_type="text_classification")
        mock_login.assert_called_once()


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch("dataquality.core.init.login", side_effect=mocked_login)
def test_init_no_token_login_full(
    mock_login: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    set_test_config(token=None, current_project_id=None, current_run_id=None)
    # When no token is passed in we should call login
    dataquality.init(task_type="text_classification")
    mock_login.assert_called_once()
    # We also test the remaining init flow
    assert config.current_run_id
    assert config.current_project_id


@patch.object(
    dataquality.core.init.ApiClient, "get_current_user", side_effect=GalileoException
)
@patch("dataquality.core.init.login", side_effect=LoginInvoked)
def test_init_expired_token_login(
    mock_login: MagicMock, mock_current_user: MagicMock, set_test_config: Callable
) -> None:
    # When a token is passed in but user auth fails we should call login
    with pytest.raises(LoginInvoked):
        dataquality.init(task_type="text_classification")
        mock_login.assert_called_once()


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(
    dataquality.core.init.ApiClient, "get_current_user", side_effect=GalileoException
)
@patch("dataquality.core.init.login", side_effect=mocked_login)
def test_init_expired_token_login_full(
    mock_login: MagicMock,
    mock_current_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    set_test_config(current_project_id=None, current_run_id=None)
    # When a token is passed in but user auth fails we should call login
    dataquality.init(task_type="text_classification")
    mock_login.assert_called_once()
    # We also test the remaining init flow
    assert config.current_run_id
    assert config.current_project_id


@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=False)
@patch("dataquality.core.init.login", side_effect=LoginInvoked)
def test_init_invalid_user_login(
    mock_login: MagicMock, mock_valid_user: MagicMock
) -> None:
    # When current user is not valid we should call login
    with pytest.raises(LoginInvoked):
        dataquality.init(task_type="text_classification")
        mock_login.assert_called_once()


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=False)
@patch("dataquality.core.init.login", side_effect=mocked_login)
def test_init_invalid_user_login_full(
    mock_login: MagicMock,
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    set_test_config: Callable,
) -> None:
    set_test_config(current_project_id=None, current_run_id=None)
    # When current user is not valid we should call login
    dataquality.init(task_type="text_classification")
    mock_login.assert_called_once()
    # We also test the remaining init flow
    assert config.current_run_id
    assert config.current_project_id


@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
def test_init_bad_task(
    mock_valid_user: MagicMock, mock_requests_get: MagicMock
) -> None:
    with pytest.raises(GalileoException):
        dataquality.init(task_type="not_text_classification")


@patch("dataquality.login")
def test_reconfigure_sets_env_vars(mock_login: MagicMock) -> None:
    os.environ["GALILEO_CONSOLE_URL"] = "https://console.fakecompany.io"
    dataquality.configure()
    assert dataquality.config.api_url == config.api_url == "https://api.fakecompany.io"
    os.environ["GALILEO_CONSOLE_URL"] = "https://console.newfake.de"
    dataquality.configure()
    assert dataquality.config.api_url == config.api_url == "https://api.newfake.de"

    assert mock_login.call_count == 2


@patch("requests.post", side_effect=mocked_login_requests)
@patch("requests.get", side_effect=mocked_login_requests)
def test_reconfigure_resets_user_token(
    mock_get_request: MagicMock,
    mock_post_request: MagicMock,
    set_test_config: Callable,
) -> None:
    set_test_config(token="old_token")

    os.environ[GALILEO_AUTH_METHOD] = "email"
    os.environ["GALILEO_USERNAME"] = "user"
    os.environ["GALILEO_PASSWORD"] = "password"
    dataquality.configure()
    assert all([config.token == "mock_token", config.token != "old_token"])


@patch("dataquality.login", side_effect=mocked_login)
def test_reconfigure_resets_user_token_login_mocked(
    mock_login: MagicMock, set_test_config: Callable
) -> None:
    set_test_config(token="old_token")
    dataquality.configure()
    assert all([config.token == "mock_token", config.token != "old_token"])
    mock_login.assert_called_once()


@patch("requests.post", side_effect=mocked_create_project_run)
@patch("requests.get", side_effect=mocked_get_project_run)
@patch.object(dataquality.core.init.ApiClient, "valid_current_user", return_value=True)
@pytest.mark.parametrize(
    "run_name,exc",
    [
        ("my-run-#1", True),
        ("$run!$", True),
        ("a-bad-run##", True),
        ("A good run1", False),
        ("a-good-run-123", False),
        ("MY Run_1-2022", False),
    ],
)
def test_bad_names(
    mock_valid_user: MagicMock,
    mock_requests_get: MagicMock,
    mock_requests_post: MagicMock,
    run_name: str,
    exc: bool,
    set_test_config: Callable,
):
    init = partial(
        dataquality.init, task_type="text_classification", project_name="proj"
    )
    if exc:
        with pytest.raises(GalileoException):
            init(run_name=run_name)
    else:
        init(run_name=run_name)



from enum import Enum, unique


@unique
class NERErrorType(str, Enum):
    wrong_tag = "wrong_tag"
    missed_label = "missed_label"
    span_shift = "span_shift"
    ghost_span = "ghost_span"
    none = "None"  # Indicating that there is no error for this span (correct pred)


@unique
class TaggingSchema(str, Enum):
    BIO = "BIO"
    BILOU = "BILOU"
    BIOES = "BIOES"
    # IOB2 = "IOB2"
    # IOB = "IOB"
    # BILOES = "BILOES"


class NERColumns(str, Enum):
    id = "id"
    sample_id = "sample_id"
    split = "split"  # type: ignore
    epoch = "epoch"
    is_gold = "is_gold"
    is_pred = "is_pred"
    span_start = "span_start"
    span_end = "span_end"
    gold = "gold"
    pred = "pred"
    data_error_potential = "data_error_potential"
    galileo_error_type = "galileo_error_type"
    emb = "emb"



import os
from typing import Callable
from unittest.mock import MagicMock, patch

import pytest
import requests

from dataquality.core._config import set_config, url_is_localhost
from dataquality.exceptions import GalileoException


def test_console_url(set_test_config: Callable) -> None:
    os.environ["GALILEO_CONSOLE_URL"] = "https://console.mytest.rungalileo.io"
    cfg = set_config()
    assert cfg.api_url == "https://api.mytest.rungalileo.io"


@pytest.mark.parametrize(
    "console_url",
    ["http://localhost:3000", "http://127.0.0.1:3000"],
)
def test_console_url_local(set_test_config: Callable, console_url: str) -> None:
    os.environ["GALILEO_CONSOLE_URL"] = console_url
    cfg = set_config()
    assert cfg.api_url == "http://localhost:8088"


def test_bad_console_url(set_test_config: Callable) -> None:
    """If console is not in the console url, dont use it"""
    os.environ["GALILEO_CONSOLE_URL"] = "https://something.mytest2.rungalileo.io"
    cfg = set_config()
    assert cfg.api_url != "https://api.mytest2.rungalileo.io"


@pytest.mark.parametrize(
    "url,expected",
    [
        ("https://console.domain.com", False),
        ("http://localhost:3000", True),
        ("http://127.0.0.1:8088", True),
    ],
)
def test_url_is_localhost(url: str, expected: bool) -> bool:
    assert url_is_localhost(url) is expected


@patch("dataquality.core._config.requests.get")
@pytest.mark.noautofixt
def test_validate_bad_config_url(get_mock: MagicMock) -> None:
    get_mock.side_effect = requests.ConnectionError("No connection!")
    bad_url = "https://console.mytest2.rungalileo.io/badurl"
    os.environ["GALILEO_CONSOLE_URL"] = bad_url
    with pytest.raises(GalileoException) as e:
        set_config()
    assert e.value.args[0].startswith(f"The provided console URL {bad_url} is invalid")


def test_handle_extra_slash_in_console(set_test_config: Callable) -> None:
    """Removes slashes in an otherwise valid console url"""
    os.environ["GALILEO_CONSOLE_URL"] = "https://console.mytest2.rungalileo.io///"
    cfg = set_config()
    assert cfg.api_url == "https://api.mytest2.rungalileo.io"



import numpy as np

from dataquality.utils.spacy_integration import convert_spacy_ner_logits_to_valid_logits


def test_convert_spacy_ner_logits_to_valid_logits():
    arr_1 = np.array([0.9, 0.8, 0.7, 0.6, 0.5, 0.4])
    pred_1 = 2

    assert np.allclose(
        convert_spacy_ner_logits_to_valid_logits(arr_1, pred_1),
        [-0.6, -0.6, 0.7, 0.6, 0.5, 0.4],
    )

    arr_2 = np.array([3.4, -0.9, 0.4, 2.8])
    pred_2 = 0

    assert np.allclose(
        convert_spacy_ner_logits_to_valid_logits(arr_2, pred_2), [3.4, -0.9, 0.4, 2.8]
    )

    arr_3 = np.array([3.4, -0.9, 0.4, 2.8])
    pred_3 = 1

    assert np.allclose(
        convert_spacy_ner_logits_to_valid_logits(arr_3, pred_3),
        [-1.9, -0.9, -1.9, -1.9],
    )

    print("pausing")



import warnings
from typing import Any, DefaultDict, Dict, List, Optional

import numpy as np
import pandas as pd
import vaex
from vaex.dataframe import DataFrame

from dataquality.core._config import config
from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger.base_data_logger import MetaType
from dataquality.loggers.data_logger.text_classification import (
    MetasType,
    TextClassificationDataLogger,
)
from dataquality.loggers.logger_config.text_multi_label import (
    TextMultiLabelLoggerConfig,
    text_multi_label_logger_config,
)
from dataquality.schemas import __data_schema_version__
from dataquality.schemas.split import Split

DATA_FOLDERS = ["emb", "prob", "data"]


class TextMultiLabelDataLogger(TextClassificationDataLogger):
    """
    Class for logging input data/metadata of Text Multi Label models to Galileo.

    * texts: The raw text inputs for model training. List[str]
    * task_labels: the list of ground truth labels aligned to each text field. Each text
    field input must have the same number of labels (which must be the number of tasks)
    List[List[str]]
    * ids: Optional unique indexes for each record. If not provided, will default to
    the index of the record. Optional[List[int]]
    * split: The split for training/test/validation

    .. code-block:: python

        task_labels = [["A", "B", "C"], ["Foo", "Bar"], ["Apple", "Orange", "Grape"]]
        tasks = ["Task_0", "Task_1", "Task_2"]
        dq.init("text_multi_label")
        dq.set_tasks_for_run(tasks)
        dq.set_labels_for_run(labels = task_labels)

        texts: List[str] = [
            "Text sample 1",
            "Text sample 2",
            "Text sample 3",
        ]

        task_labels: List[str] = [
            ["A", "Foo", "Grape"],
            ["C", "Foo", "Apple"],
            ["B", "Bar", "Orange"]
        ]

        ids: List[int] = [0, 1, 2]
        meta = {"sample_quality": [5.3, 9.1, 2.7]}
        split = "training"

        dq.log_data_samples(
            texts=texts, task_labels=task_labels, ids=ids, meta=meta, split=split
        )
    """

    __logger_name__ = "text_multi_label"
    logger_config: TextMultiLabelLoggerConfig = (
        text_multi_label_logger_config  # type: ignore
    )

    def __init__(
        self,
        texts: List[str] = None,
        labels: List[List[str]] = None,
        ids: List[int] = None,
        split: str = None,
        meta: MetasType = None,
    ) -> None:
        """Create data logger.

        :param text: The raw text inputs for model training. List[str]
        :param labels: the ground truth labels aligned to each text field.
        List[List[str]]
        :param ids: Optional unique indexes for each record. If not provided, will
        default to the index of the record. Optional[List[Union[int,str]]]
        :param split: The split for training/test/validation
        """
        super().__init__(texts=texts, ids=ids, split=split, meta=meta)
        if labels is not None:
            self.labels: List[List[str]] = [
                [str(i) for i in tl] for tl in labels  # type: ignore
            ]
        else:
            self.labels = []

    def log_data_sample(
        self,
        *,
        text: str,
        id: int,
        label: Optional[str] = None,
        split: Optional[Split] = None,
        inference_name: Optional[str] = None,
        meta: Optional[MetaType] = None,
        task_labels: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Log a single input sample for text multi-label
        :param text: str the text sample
        :param id: The sample ID
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param meta: Dict[str, Union[str, int, float]]. Metadata for the text sample
            Format is the {"metadata_field_name": metadata_field_value}
        :param task_labels: List[str] The label of each task for this sample
            Required if split is not inference
        """
        self.validate_kwargs(kwargs)
        if label:
            raise GalileoException("In multi-label, use task_labels instead of label")

        task_labels = [str(i) for i in task_labels] if task_labels else []

        self.texts = [text]
        self.ids = [id]
        self.split = split
        self.labels = [task_labels] if task_labels else []
        self.inference_name = inference_name
        self.meta = {i: [meta[i]] for i in meta} if meta else {}
        self.log()

    def log_data_samples(
        self,
        *,
        texts: List[str],
        ids: List[int],
        labels: Optional[List[str]] = None,
        split: Optional[Split] = None,
        inference_name: str = None,
        meta: MetasType = None,
        task_labels: Optional[List[List[str]]] = None,
        **kwargs: Any,
    ) -> None:
        """Log input samples for text multi-label

        :param texts: List[str] text samples
        :param ids: List[int,str] IDs for each text sample
        :param split: train/test/validation/inference. Can be set here or via
            dq.set_split
        :param meta: Dict[str, List[str, int, float]]. Metadata for each text sample
            Format is the {"metadata_field_name": [metdata value per sample]}
        :param task_labels: List[List[str]] list of labels for each task for each
            text sample. Required if not in inference
        """
        self.validate_kwargs(kwargs)
        if labels is not None:
            raise GalileoException("In multi-label, use task_labels instead of labels")
        self.texts = texts
        self.ids = ids
        self.split = split
        self.meta = meta or {}
        if task_labels is not None:
            self.labels = [[str(i) for i in tl] for tl in task_labels]
        else:
            self.labels = []
        self.log()

    def _process_label(self, batches: DefaultDict, label: Any) -> DefaultDict:
        """In multi-label, label will be a list of strings instead of a string"""
        # In binary multi-label, it will be a single string
        if self.logger_config.binary:
            return super()._process_label(batches, label)
        batches["label"].append(self._convert_tensor_ndarray(label).tolist())
        return batches

    def _log_dict(
        self, d: Dict, meta: Dict, split: Split = None, inference_name: str = None
    ) -> None:
        self.log_data_samples(
            texts=d["text"],
            task_labels=d["label"],
            ids=d["id"],
            split=split,
            inference_name=inference_name,
            meta=meta,
        )

    def _process_binary_labels(self) -> None:
        """In binary multi labels, users will simply log the samples that are
        active in this case. If tasks are ["A","B","C"] and sample 1 ground truth is
        ["A"], sample 2 is ["A","C"] the user will log
        [
            ("Sample 1", ["A"]),
            ("Sample 1", ["A","C"])
        ]

        We want to convert that to
        [
            ("Sample 1", ["A", "NOT_B", "NOT_C"]),
            ("Sample 1", ["A","NOT_B", "C"])
        ]
        """
        # Assert for mypy
        assert self.logger_config.tasks, (
            "You must call dq.set_tasks_for_run(..., binary=True) before logging"
            " in binary multi-label"
        )
        clean_task_labels = []
        for sample_task_labels in self.labels:
            clean_sample_labels = []
            sample_label_set = set(sample_task_labels)
            for task in self.logger_config.tasks:
                if task in sample_label_set:
                    clean_sample_labels.append(task)
                else:
                    clean_sample_labels.append(f"NOT_{task}")
            clean_task_labels.append(clean_sample_labels)
        self.labels = clean_task_labels

    def validate(self) -> None:
        """
        Parent validation (text_classification) with additional validation on labels

        in multi_label modeling, each element in self.labels should itself be a list
        """
        if self.logger_config.binary:
            self._process_binary_labels()
        self.logger_config.observed_num_tasks = len(self.labels[0])
        super().validate()

    def validate_logged_labels(self) -> None:
        for ind, input_labels in enumerate(self.labels):
            assert isinstance(
                input_labels, (list, np.ndarray, pd.Series)
            ), "labels must be a list of lists in multi-label tasks"
            assert len(input_labels) == self.logger_config.observed_num_tasks, (
                f"Each {self.split} input must have the same number of labels. "
                f"Expected {self.logger_config.observed_num_tasks} based on record 0 "
                f"but saw {len(input_labels)} for input record {ind}. If this is a "
                f"binary multi label and you are logging the active tasks, call"
                f"dq.set_tasks_for_run(tasks, binary=True) and log again"
            )

            for task_ind, task_label in enumerate(input_labels):
                # In binary case its just the 1 label so we make it a list to conform
                self._validate_task_labels(ind, task_ind, task_label)

    def _validate_task_labels(
        self, input_ind: int, task_ind: int, task_label: str
    ) -> None:
        # Capture the observed labels per task
        self.logger_config.observed_labels[task_ind].update({task_label})
        observed_task_labels = self.logger_config.observed_labels[task_ind]
        # If the user has already set labels validate logged ones are valid
        if self.logger_config.labels:
            set_task_labels = self.logger_config.labels[task_ind]
            assert observed_task_labels.issubset(set_task_labels), (
                f"The input labels you log must be exactly the same as the "
                f"labels you set in set_labels_for_run. Input record {input_ind}: "
                f"logged labels: {observed_task_labels}, set labels for that "
                f"task (task #{task_ind}): {set_task_labels}"
            )

    def _get_input_df(self) -> DataFrame:
        inp = dict(
            id=self.ids,
            text=self.texts,
            split=self.split,
            data_schema_version=__data_schema_version__,
            **self.meta,
        )
        if self.split != Split.inference.value:
            gold_array = np.array(self.labels)
            for task_num in range(self.logger_config.observed_num_tasks):
                inp[f"gold_{task_num}"] = gold_array[:, task_num]
        return vaex.from_pandas(pd.DataFrame(inp))

    @classmethod
    def _get_prob_cols(cls) -> List[str]:
        prob_cols = [f"prob_{i}" for i in range(cls.logger_config.observed_num_tasks)]
        gold_cols = [f"gold_{i}" for i in range(cls.logger_config.observed_num_tasks)]
        return ["id"] + prob_cols + gold_cols

    @classmethod
    def validate_labels(cls) -> None:
        assert cls.logger_config.labels, (
            "You must set your config labels before calling finish. "
            "See `dataquality.set_labels_for_run`"
        )
        if not cls.logger_config.tasks:
            cls.logger_config.tasks = [
                f"task_{i}" for i in range(cls.logger_config.observed_num_tasks)
            ]
            warnings.warn(
                f"No tasks were set for this run. Setting tasks to "
                f"{cls.logger_config.tasks}"
            )

        assert len(cls.logger_config.tasks) == cls.logger_config.observed_num_tasks, (
            f"You set your task names as {cls.logger_config.tasks} "
            f"({len(cls.logger_config.tasks)} tasks but based on training, your model "
            f"has {cls.logger_config.observed_num_tasks} "
            f"tasks. Use dataquality.set_tasks_for_run to update your config tasks."
        )

        assert len(cls.logger_config.labels) == cls.logger_config.observed_num_tasks, (
            f"You set your labels to be {cls.logger_config.labels} "
            f"({len(cls.logger_config.labels)} tasks) but based on training, your "
            f"model has {cls.logger_config.observed_num_tasks} tasks. "
            f"Use dataquality.set_labels_for_run to update your config labels."
        )
        assert isinstance(cls.logger_config.observed_num_labels, list), (
            f"Is your task_type correct? The observed number of labels is "
            f"{cls.logger_config.observed_num_labels}, but this should be a list of "
            f"ints, one per task. Should task {config.task_type} be set?"
        )
        assert (
            len(cls.logger_config.observed_num_labels)
            == cls.logger_config.observed_num_tasks
        ), (
            "Something went wrong with model output logging. Based on training, the "
            f"observed number of labels per task is "
            f"{cls.logger_config.observed_num_labels} indicating "
            f"{len(cls.logger_config.observed_num_labels)} tasks, but the observed "
            f"number of tasks is only {cls.logger_config.observed_num_tasks}. Ensure "
            f"you are using the logger properly and that your task_type is "
            f"correct ({config.task_type})."
        )
        for task_num, (task, task_labels, num_task_labels) in enumerate(
            zip(
                cls.logger_config.tasks,
                cls.logger_config.labels,
                cls.logger_config.observed_num_labels,
            )
        ):
            assert isinstance(task_labels, list), (
                "In the multi-label case, your config labels should be a list of lists "
                "of strings. See dataquality.set_labels_for_run to update your config "
                "labels."
            )
            assert len(task_labels) == num_task_labels, (
                f"Task {task} is set to have {len(task_labels)} labels ({task_labels}) "
                f"but based on training, your model has {num_task_labels} labels "
                f"for that task. See dataquality.set_labels_for_run to update your "
                f"config labels"
            )
            logged_labels = cls.logger_config.observed_labels[task_num]
            assert logged_labels.issubset(task_labels), (
                f"The labels set for task #{task_num} ({task}) do not align with the "
                "observed labels during logging. Labels logged for this task during "
                f"input logging: {logged_labels} -- labels set for this task: "
                f"{task_labels}. Update the labels for this task using "
                "`dq.set_labels_for_run`"
            )

    def _get_num_labels(self, df: DataFrame) -> List[int]:
        return [len(i) for i in df[:1]["prob"].values[0]]



from typing import Callable
from unittest import mock
from unittest.mock import MagicMock

import numpy as np

import dataquality as dq
from dataquality.utils.dq_logger import dq_log_file_path


@mock.patch("dataquality.core.finish._reset_run")
@mock.patch("dataquality.core.finish._version_check")
@mock.patch.object(dq.clients.api.ApiClient, "make_request")
@mock.patch.object(dq.clients.api.ApiClient, "get_presigned_url")
@mock.patch.object(dq.clients.objectstore.ObjectStore, "download_file")
@mock.patch.object(dq.clients.objectstore.ObjectStore, "_upload_file_from_local")
def test_std_log(
    mock_upload_from_local: MagicMock,
    mock_download: MagicMock,
    mock_presigned_url: MagicMock,
    mock_finish: MagicMock,
    mock_version_check: MagicMock,
    mock_reset_run: MagicMock,
    set_test_config: Callable,
    cleanup_after_use: Callable,
    input_data: Callable,
) -> None:
    """Validate that we interrupt the main process if issues occur while logging"""
    dq.set_labels_for_run(["APPLE", "ORANGE"])
    training_data = input_data(meta={"training_meta": [1.414, 123]})
    dq.log_data_samples(**training_data)

    dq.set_split("training")
    dq.log_model_outputs(
        embs=np.random.rand(2, 100),
        logits=np.random.rand(2, 2),
        ids=[1, 2],
        epoch=0,
    )
    mock_finish.return_value = {"job_name": "test_job", "link": "link"}
    mock_presigned_url.return_value = "https://google.com"
    mock_upload_from_local.return_value = None
    dq.finish(wait=False)
    mock_download.return_value = "my-file"

    x = dq.get_dq_log_file()
    assert x == dq_log_file_path()



import warnings
from collections import defaultdict
from enum import Enum, unique
from typing import Any, Dict, List, Optional, Union

import numpy as np

from dataquality.loggers.logger_config.text_classification import (
    text_classification_logger_config,
)
from dataquality.loggers.model_logger.base_model_logger import BaseGalileoModelLogger
from dataquality.schemas import __data_schema_version__
from dataquality.schemas.split import Split
from dataquality.utils.dq_logger import get_dq_logger


@unique
class GalileoModelLoggerAttributes(str, Enum):
    embs = "embs"
    probs = "probs"
    logits = "logits"
    ids = "ids"
    # mixin restriction on str (due to "str".split(...))
    split = "split"  # type: ignore
    epoch = "epoch"
    inference_name = "inference_name"

    @staticmethod
    def get_valid() -> List[str]:
        return list(map(lambda x: x.value, GalileoModelLoggerAttributes))


class TextClassificationModelLogger(BaseGalileoModelLogger):
    """
    Class for logging model output data of Text Classification models to Galileo.

    * embs: Union[List, np.ndarray, torch.Tensor, tf.Tensor]. The Embeddings per
    text sample input. Only one embedding vector is allowed per input sample.
    the `embs` parameter can be formatted either as:
        * np.ndarray
        * torch.tensor / tf.tensor
        * A list of List[float]
        * A list of numpy arrays
        * A list of tensorflow tensors
        * A list of pytorch tensors
    * logits: Union[List, np.ndarray, torch.Tensor, tf.Tensor] outputs from
     forward pass. If provided, probs will be converted automatically and DO NOT need
     to be provided. Can be formatted either as:
        * np.ndarray
        * torch.tensor / tf.tensor
        * A list of List[float]
        * A list of numpy arrays
        * A list of tensorflow tensors
        * A list of pytorch tensors
    * probs: Deprecated - the probabilities for each output sample (use logits instead)
    * ids: Indexes of each input field: List[int]. These IDs must align with the input
    IDs for each sample input. This will be used to join them together for analysis
    by Galileo.
    * split: The model training/test/validation split for the samples being logged

    ex:
    .. code-block:: python

        dq.set_epoch(0)
        dq.set_split("train")

        embs: np.ndarray = np.random.rand(4, 768)  # 4 samples, embedding dim 768
        logits: np.ndarray = np.random.rand(4, 3)  # 4 samples, 3 classes
        ids: List[int] = [0, 1, 2, 3]

        dq.log_model_outputs(embs=embs, logits=logits, ids=ids)
    """

    __logger_name__ = "text_classification"
    logger_config = text_classification_logger_config

    def __init__(
        self,
        embs: Union[List, np.ndarray] = None,
        probs: Union[List, np.ndarray] = None,
        logits: Union[List, np.ndarray] = None,
        ids: Union[List, np.ndarray] = None,
        split: str = "",
        epoch: Optional[int] = None,
        inference_name: Optional[str] = None,
    ) -> None:
        super().__init__(
            embs=embs,
            probs=probs,
            logits=logits,
            ids=ids,
            split=split,
            epoch=epoch,
            inference_name=inference_name,
        )

    @staticmethod
    def get_valid_attributes() -> List[str]:
        """
        Returns a list of valid attributes that this logger accepts
        :return: List[str]
        """
        return GalileoModelLoggerAttributes.get_valid()

    def validate(self) -> None:
        """
        Validates that the current config is correct.
        * embs, probs, and ids must exist and be the same length
        :return:
        """
        get_dq_logger().info("Handling logits and probs", split=self.split)
        if len(self.logits):
            self.logits = self._convert_tensor_ndarray(self.logits, "Prob")
            self.probs = self.convert_logits_to_probs(self.logits)
            del self.logits
        elif len(self.probs):
            warnings.warn("Usage of probs is deprecated, use logits instead")
            self.probs = self._convert_tensor_ndarray(self.probs, "Prob")

        embs_len = len(self.embs)
        probs_len = len(self.probs)
        ids_len = len(self.ids)

        get_dq_logger().info("Converting inputs to numpy arrays", split=self.split)
        self.embs = self._convert_tensor_ndarray(self.embs, "Embedding")
        self.ids = self._convert_tensor_ndarray(self.ids)

        get_dq_logger().info("Validating embedding shape", split=self.split)
        assert self.embs.ndim == 2, "Only one embedding vector is allowed per input."

        assert embs_len and probs_len and ids_len, (
            f"All of emb, probs, and ids for your logger must be set, but "
            f"got emb:{bool(embs_len)}, probs:{bool(probs_len)}, ids:{bool(ids_len)}"
        )

        assert embs_len == probs_len == ids_len, (
            f"All of emb, probs, and ids for your logger must be the same "
            f"length, but got (emb, probs, ids) -> ({embs_len},{probs_len}, {ids_len})"
        )

        # User may manually pass in 'train' instead of 'training' / 'test' vs 'testing'
        # but we want it to conform
        try:
            self.split = Split[self.split].value
        except KeyError:
            get_dq_logger().error("Provided a bad split", split=self.split)
            raise AssertionError(
                f"Split should be one of {Split.get_valid_attributes()} "
                f"but got {self.split}"
            )

        if self.epoch:
            assert isinstance(self.epoch, int), (
                f"If set, epoch must be int but was " f"{type(self.epoch)}"
            )
            if self.epoch > self.logger_config.last_epoch:
                self.logger_config.last_epoch = self.epoch

    def write_model_output(self, model_output: Dict) -> None:
        self._set_num_labels(model_output)
        super().write_model_output(model_output)

    def _get_data_dict(self) -> Dict[str, Any]:
        data = defaultdict(list)
        for record_id, prob, emb in zip(self.ids, self.probs, self.embs):
            # Handle binary classification by making it 2-class classification
            p = [1 - prob[0], prob[0]] if len(prob) == 1 else prob
            record = {
                "id": record_id,
                "epoch": self.epoch,
                "split": Split[self.split].value,
                "emb": emb,
                "prob": p,
                "pred": int(np.argmax(prob)),
                "data_schema_version": __data_schema_version__,
            }
            if self.split == Split.inference:
                record.update(inference_name=self.inference_name)
            for k in record.keys():
                data[k].append(record[k])
        return data

    def _set_num_labels(self, data: Dict) -> None:
        self.logger_config.observed_num_labels = len(data["prob"][0])

    def __setattr__(self, key: Any, value: Any) -> None:
        if key not in self.get_valid_attributes():
            raise AttributeError(
                f"{key} is not a valid attribute of {self.__logger_name__} logger. "
                f"Only {self.get_valid_attributes()}"
            )
        super().__setattr__(key, value)



from collections import defaultdict
from typing import DefaultDict, Dict, List

import numpy as np
import spacy
from spacy.tokens import Doc
from spacy.training import offsets_to_biluo_tags

from dataquality.exceptions import GalileoException


def validate_obj(an_object: object, check_type: type, has_attr: str) -> None:
    if not isinstance(an_object, check_type):
        raise GalileoException(
            f"Expected a {check_type}. Received {str(type(an_object))}"
        )

    if not hasattr(an_object, has_attr):
        raise GalileoException(f"Your {check_type} must have a {has_attr} attribute")


def convert_spacy_ner_logits_to_valid_logits(
    logits: np.ndarray, pred: int
) -> np.ndarray:
    """Converts ParserStepModel per token logits to their matching valid logits.

    Not all logits outputted by the spacy model are valid logits, for this reason
    spacy will ignore potential actions even if they might've had the largest prob mass.
    To account for this, we first sort the logits for each token and then zero out
    all logits larger than the predicted logit (as these must've been ignored by spacy
    or else they would've become the prediction).

    :param logits: ParserStepModel logits for a single token, minus the -U tag logit
    shape of [num_classes]
    :param pred: the idx of the spacy's valid prediction from the logits
    :return: np array of logits. shape of [num_classes]
    """
    assert len(logits.shape) == 1
    logits = logits.copy()

    # we considered first -inf, but -inf skews the softmax'd probs much more than this
    zeroing_value = logits.min() - 1
    # Sort in descending order
    argsorted_sample_logits = np.flip(np.argsort(logits))

    # Get all logit indices where pred_logit > logit
    # These are 'valid' because spacy ignored all logits > pred_logit
    # as it they were determined to not be possible given the current state.
    valid_logit_indices = argsorted_sample_logits[
        np.where(argsorted_sample_logits == pred)[0][0] :
    ]

    # non valid_logit_indices should be set to 0
    zero_out_mask = np.ones(logits.shape, bool)
    zero_out_mask[valid_logit_indices] = False
    logits[zero_out_mask] = zeroing_value
    return logits


def convert_spacy_ents_for_doc_to_predictions(
    docs: Dict[int, Doc], labels: List[str]
) -> DefaultDict[int, List[int]]:
    """Converts spacy's representation of ner spans to their per token predictions.

    Uses some spacy utility code to convert from start/end/label representation to the
    BILUO per token corresponding tagging scheme.

    """
    prediction_indices = defaultdict(list)
    for doc_id, doc in docs.items():
        pred_output = offsets_to_biluo_tags(
            doc, [(ent.start_char, ent.end_char, ent.label_) for ent in doc.ents]
        )
        pred_output_ind = [labels.index(tok_pred) for tok_pred in pred_output]
        prediction_indices[doc_id] = pred_output_ind
    return prediction_indices


def validate_spacy_version() -> None:
    """Validates the user is on a version of spacy we support"""
    if spacy.__version__ != "3.2.1":
        raise GalileoException(
            "Currently we only support watching SpaCy models running version 3.2.1 of "
            f"SpaCy. You have version {spacy.__version__}. Please install 3.2.1 using "
            f"the following: 'pip install --upgrade spacy==3.2.1' and then restart your"
            f"IPython kernel"
        )



import os
import warnings
from typing import Dict, List, Optional, Union
from uuid import uuid4

import vaex
from vaex.dataframe import DataFrame

from dataquality.clients.api import ApiClient
from dataquality.clients.objectstore import ObjectStore
from dataquality.exceptions import GalileoException, GalileoWarning
from dataquality.schemas.dataframe import FileType
from dataquality.schemas.metrics import FilterParams
from dataquality.schemas.ner import TaggingSchema
from dataquality.schemas.split import Split
from dataquality.schemas.task_type import TaskType

api_client = ApiClient()
object_store = ObjectStore()


def get_run_summary(
    project_name: str,
    run_name: str,
    split: Split,
    task: Optional[str] = None,
    inference_name: Optional[str] = None,
    filter: Union[FilterParams, Dict] = None,
) -> Dict:
    """Gets the summary for a run/split

    Calculates metrics (f1, recall, precision) overall (weighted) and per label.
    Also returns the top 50 rows of the dataframe (sorted by data_error_potential)

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param task: (If multi-label only) the task name in question
    :param inference_name: (If inference split only) The inference split name
    :param filter: Optional filter to provide to restrict the summary to only to
        matching rows. See `dq.schemas.metrics.FilterParams`
    """
    filter_params = _validate_filter(filter)
    return api_client.get_run_summary(
        project_name, run_name, split, task, inference_name, filter_params=filter_params
    )


def get_metrics(
    project_name: str,
    run_name: str,
    split: Split,
    task: Optional[str] = None,
    inference_name: Optional[str] = None,
    category: str = "gold",
    filter: Union[FilterParams, Dict] = None,
) -> Dict[str, List]:
    """Calculates available metrics for a run/split, grouped by a particular category

    The category/column provided (can be gold, pred, or any categorical metadata column)
    will result in metrics per "group" or unique value of that category/column

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param task: (If multi-label only) the task name in question
    :param inference_name: (If inference split only) The inference split name
    :param category: The category/column to calculate metrics for. Default "gold"
        Can be "gold" for ground truth, "pred" for predicted values, or any metadata
        column logged (or smart feature).
    :param filter: Optional filter to provide to restrict the metrics to only to
        matching rows. See `dq.schemas.metrics.FilterParams`
    """
    metrics = api_client.get_run_metrics(
        project_name,
        run_name,
        split,
        task=task,
        inference_name=inference_name,
        category=category,
        filter_params=_validate_filter(filter),
    )
    # Filter out metrics not available for this request
    metrics = {k: v for k, v in metrics.items() if v}
    return metrics


def display_distribution(
    project_name: str,
    run_name: str,
    split: Split,
    task: Optional[str] = None,
    inference_name: Optional[str] = None,
    column: str = "data_error_potential",
    filter: Union[FilterParams, Dict] = None,
) -> None:
    """Displays the column distribution for a run. Plotly must be installed

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param task: (If multi-label only) the task name in question
    :param inference_name: (If inference split only) The inference split name
    :param column: The column to get the distribution for. Default data error potential
    :param filter: Optional filter to provide to restrict the distribution to only to
        matching rows. See `dq.schemas.metrics.FilterParams`
    """
    try:
        import plotly.express as px
    except ImportError:
        raise GalileoException(
            "You must install plotly to use this function. Run `pip install plotly`"
        )
    distribution = api_client.get_column_distribution(
        project_name,
        run_name,
        split,
        column=column,
        task=task,
        inference_name=inference_name,
        filter_params=_validate_filter(filter),
    )
    bins, counts = distribution["bins"], distribution["counts"]
    labels = {"x": column, "y": "Count"}

    color_scale, color = None, None
    if column == "data_error_potential":
        summary = api_client.get_run_summary(
            project_name, run_name, split, task, inference_name
        )["split_run_results"]
        easy = summary["easy_samples_threshold"]
        hard = summary["hard_samples_threshold"]
        color_scale = [
            (0, "green"),
            (easy, "yellow"),
            (hard, "red"),
            (1, "darkred"),
        ]
        color = bins[1:]

    fig = px.bar(
        x=bins[1:],
        y=counts,
        labels=labels,
        color=color,
        color_continuous_scale=color_scale,
    )
    fig.show()


def get_dataframe(
    project_name: str,
    run_name: str,
    split: Split,
    inference_name: str = "",
    file_type: FileType = FileType.arrow,
    include_embs: bool = False,
    include_probs: bool = False,
    include_token_indices: bool = False,
    hf_format: bool = False,
    tagging_schema: Optional[TaggingSchema] = None,
    filter: Union[FilterParams, Dict] = None,
) -> DataFrame:
    """Gets the dataframe for a run/split

    Downloads an arrow (or specified type) file to your machine and returns a loaded
    Vaex dataframe.

    Special note for NER. By default, the data will be downloaded at a sample level
    (1 row per sample text), with spans for each sample in a `spans` column in a
    spacy-compatible JSON format. If include_emb is True, the data will be expanded
    into span level (1 row per span, with sample text repeated for each span row), in
    order to join the span-level embeddings

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param inference_name: Required if split is inference. The name of the inference
        split to get data for.
    :param file_type: The file type to download the data as. Default arrow
    :param include_embs: Whether to include the embeddings in the data. Default False
    :param include_probs: Whether to include the probs in the data. Default False
    :param include_token_indices: (NER only) Whether to include logged
        text_token_indices in the data. Useful for reconstructing tokens for retraining
    :param hf_format: (NER only)
        Whether to export the dataframe in a HuggingFace compatible format
    :param tagging_schema: (NER only)
        If hf_format is True, you must pass a tagging schema
    :param filter: Optional filter to provide to restrict the distribution to only to
        matching rows. See `dq.schemas.metrics.FilterParams`
    """
    project_id, run_id = api_client._get_project_run_id(project_name, run_name)
    task_type = api_client.get_task_type(project_id, run_id)

    file_name = f"/tmp/{uuid4()}-data.{file_type}"
    filter_params = _validate_filter(filter)
    api_client.export_run(
        project_name,
        run_name,
        split,
        inference_name=inference_name,
        file_name=file_name,
        filter_params=filter_params,
        hf_format=hf_format,
        tagging_schema=tagging_schema,
    )
    data_df = vaex.open(file_name)
    return _process_exported_dataframe(
        data_df,
        project_name,
        run_name,
        split,
        task_type,
        inference_name,
        include_embs,
        include_probs,
        include_token_indices,
        hf_format,
    )


def get_edited_dataframe(
    project_name: str,
    run_name: str,
    split: Split,
    inference_name: str = "",
    file_type: FileType = FileType.arrow,
    include_embs: bool = False,
    include_probs: bool = False,
    include_token_indices: bool = False,
    hf_format: bool = False,
    tagging_schema: Optional[TaggingSchema] = None,
) -> DataFrame:
    """Gets the edited dataframe for a run/split

    Exports a run/split's data with all active edits in the edits cart and returns
    a vaex dataframe

    Special note for NER. By default, the data will be downloaded at a sample level
    (1 row per sample text), with spans for each sample in a `spans` column in a
    spacy-compatible JSON format. If include_emb is True, the data will be expanded
    into span level (1 row per span, with sample text repeated for each span row), in
    order to join the span-level embeddings

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param inference_name: Required if split is inference. The name of the inference
        split to get data for.
    :param file_type: The file type to download the data as. Default arrow
    :param include_embs: Whether to include the embeddings in the data. Default False
    :param include_probs: Whether to include the probs in the data. Default False
    :param include_token_indices: (NER only) Whether to include logged
        text_token_indices in the data. Useful for reconstructing tokens for retraining
    :param hf_format: (NER only)
        Whether to export the dataframe in a HuggingFace compatible format
    :param tagging_schema: (NER only)
        If hf_format is True, you must pass a tagging schema
    """
    project_id, run_id = api_client._get_project_run_id(project_name, run_name)
    task_type = api_client.get_task_type(project_id, run_id)

    file_name = f"/tmp/{uuid4()}-data.{file_type}"
    api_client.export_edits(
        project_name,
        run_name,
        split,
        inference_name=inference_name,
        file_name=file_name,
        hf_format=hf_format,
        tagging_schema=tagging_schema,
    )
    data_df = vaex.open(file_name)
    return _process_exported_dataframe(
        data_df,
        project_name,
        run_name,
        split,
        task_type,
        inference_name,
        include_embs,
        include_probs,
        include_token_indices,
        hf_format,
    )


def _process_exported_dataframe(
    data_df: DataFrame,
    project_name: str,
    run_name: str,
    split: Split,
    task_type: TaskType,
    inference_name: str = "",
    include_embs: bool = False,
    include_probs: bool = False,
    include_token_indices: bool = False,
    hf_format: bool = False,
) -> DataFrame:
    """Process dataframe after export of run or edits.

    See `get_dataframe` and `get_edited_dataframe` for details"""
    # See docstring. In this case, we need span-level data
    # You can't attach embeddings to the huggingface data, since the HF format is
    # sample level, and the embeddings are span level
    if include_embs and task_type == TaskType.text_ner and not hf_format:
        # In NER, the `probabilities` contains the span level data
        span_df = get_probabilities(project_name, run_name, split, inference_name)
        # These are the token (not char) indices, lets make that clear
        span_df.rename("span_start", "span_token_start")
        span_df.rename("span_end", "span_token_end")
        data_df = data_df[["sample_id", "text", "spans"]].join(
            span_df, on="sample_id", allow_duplication=True
        )

    tasks = []
    if task_type == TaskType.text_multi_label:
        tasks = api_client.get_tasks_for_run(project_name, run_name)
        labels = [
            api_client.get_labels_for_run(project_name, run_name, task)
            for task in tasks
        ]
        data_df = _index_df(data_df, labels, tasks)
        data_df = _clean_mltc_df(data_df)
    if task_type == TaskType.text_classification:
        labels_per_task = api_client.get_labels_for_run(project_name, run_name)
        data_df = _index_df(data_df, labels_per_task)

    if include_embs:
        # Embeddings are span level, but huggingface is sample level, so can't combine
        if hf_format:
            warnings.warn(
                "Embeddings are not available in HF format, ignoring", GalileoWarning
            )
        else:
            emb_df = get_embeddings(project_name, run_name, split, inference_name)
            data_df = data_df.join(emb_df, on="id")
    if include_probs:
        if task_type == task_type.text_ner:
            warnings.warn(
                "Probabilities are not available for NER runs, ignoring", GalileoWarning
            )
        elif hf_format:
            warnings.warn(
                "Probabilities are not available in HF format, ignoring", GalileoWarning
            )
        else:
            prob_df = get_probabilities(project_name, run_name, split, inference_name)
            prob_cols = prob_df.get_column_names(regex="prob*") + ["id"]
            data_df = data_df.join(prob_df[prob_cols], on="id")
            data_df = _rename_prob_cols(data_df, tasks)
    if include_token_indices:
        if task_type != task_type.text_ner:
            warnings.warn(
                "Token indices are only available for NER, ignoring", GalileoWarning
            )
        else:
            raw_tokens = get_raw_data(project_name, run_name, split)
            raw_tokens = raw_tokens[["id", "text_token_indices"]]
            if "sample_id" in data_df.get_column_names():
                data_df.rename("sample_id", "id")
            data_df = data_df.join(raw_tokens, on="id")
    return data_df


def get_epochs(project_name: str, run_name: str, split: Split) -> List[int]:
    """Returns the epochs logged for a run/split

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    """
    return api_client.get_epochs_for_run(project_name, run_name, split)


def get_embeddings(
    project_name: str,
    run_name: str,
    split: Split,
    inference_name: str = "",
    epoch: int = None,
) -> DataFrame:
    """Downloads the embeddings for a run/split at an epoch as a Vaex dataframe.

    If not provided, will take the embeddings from the final epoch. Note that only the
    n and n-1 epoch embeddings are available for download

    An hdf5 file will be downloaded to local and a Vaex dataframe will be returned

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param inference_name: Required if split is inference
    :param epoch: The epoch to get embeddings for. Default final epoch
    """
    return _get_hdf5_file_for_epoch(
        project_name,
        run_name,
        split,
        "emb/emb.hdf5",
        inference_name,
        epoch,
    )


def get_probabilities(
    project_name: str,
    run_name: str,
    split: Split,
    inference_name: str = "",
    epoch: int = None,
) -> DataFrame:
    """Downloads the probabilities for a run/split at an epoch as a Vaex dataframe.

    If not provided, will take the probabilities from the final epoch.

    An hdf5 file will be downloaded to local and a Vaex dataframe will be returned

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param inference_name: Required if split is inference
    :param epoch: The epoch to get embeddings for. Default final epoch
    """
    return _get_hdf5_file_for_epoch(
        project_name, run_name, split, "prob/prob.hdf5", inference_name, epoch
    )


def get_raw_data(
    project_name: str,
    run_name: str,
    split: Split,
    inference_name: str = "",
    epoch: int = None,
) -> DataFrame:
    """Downloads the raw logged data for a run/split at an epoch as a Vaex dataframe.

    If not provided, will take the probabilities from the final epoch.

    An hdf5 file will be downloaded to local and a Vaex dataframe will be returned

    :param project_name: The project name
    :param run_name: The run name
    :param split: The split (training/test/validation/inference)
    :param inference_name: Required if split is inference
    :param epoch: The epoch to get embeddings for. Default final epoch
    """

    project_id, run_id = api_client._get_project_run_id(project_name, run_name)
    task_type = api_client.get_task_type(project_id, run_id)
    if task_type == TaskType.text_ner:
        object_name = "data/data.arrow"
    else:
        object_name = "data/data.hdf5"
    return _get_hdf5_file_for_epoch(
        project_name, run_name, split, object_name, inference_name, epoch
    )


def get_xray_cards(
    project_name: str, run_name: str, split: Split, inference_name: Optional[str] = None
) -> List[Dict[str, str]]:
    """Get xray cards for a project/run/split

    Xray cards are automatic insights calculated and provided by Galileo on your data
    """
    return api_client.get_xray_cards(project_name, run_name, split, inference_name)


def get_labels_for_run(
    project_name: str, run_name: str, task: Optional[str] = None
) -> List[str]:
    """Gets labels for a given run. If multi-label, a task must be provided"""
    return api_client.get_labels_for_run(project_name, run_name, task)


def get_tasks_for_run(project_name: str, run_name: str) -> List[str]:
    """Gets task names for a multi-label run"""
    return api_client.get_tasks_for_run(project_name, run_name)


def _get_hdf5_file_for_epoch(
    project_name: str,
    run_name: str,
    split: Split,
    object_name: str,
    inference_name: str = "",
    epoch: int = None,
) -> DataFrame:
    emb = "emb" in object_name
    if split == Split.inference and inference_name:
        split_path = inference_name
    else:
        split_path = str(_validate_epoch(project_name, run_name, split, epoch, emb=emb))
    project_id, run_id = api_client._get_project_run_id(project_name, run_name)
    object_name = f"{project_id}/{run_id}/{split}/{split_path}/{object_name}"
    file_name = f"/tmp/{uuid4()}-{os.path.split(object_name)[-1]}"
    object_store.download_file(object_name, file_name)
    return vaex.open(file_name)


def _validate_epoch(
    project_name: str, run_name: str, split: Split, epoch: int = None, emb: bool = False
) -> int:
    epochs = get_epochs(project_name, run_name, split)
    last_epoch = epochs[-1]
    if not epochs:
        raise GalileoException(f"No epochs found for {project_name}/{run_name}")
    if epoch is None:
        epoch = last_epoch
    if epoch not in epochs:
        raise GalileoException(
            f"Run {project_name}/{run_name} has epochs {epochs}. You requested {epoch}"
        )
    # We only log the last 2 epochs of embeddings, so it must be one of those
    if emb and epoch < last_epoch - 1:
        raise GalileoException(
            f"Only the last 2 epochs of embeddings are available. "
            f"Must request {last_epoch} or {last_epoch-1}"
        )
    return epoch


def _index_df(df: DataFrame, labels: List, tasks: Optional[List] = None) -> DataFrame:
    """Indexes gold and pred columns"""
    # We do this so if this is TC (no tasks), we can still iterate with the same logic
    tasks = tasks or [None]
    for ind, task in enumerate(tasks):
        # If multi label, must do it per task. If TC, then it's just 1 list of labels
        task_labels = labels[ind] if task else labels
        for col in ["gold", "pred"]:
            if col not in df.get_column_names():
                continue
            df_col = f"{col}_{task}" if task else col
            df[f"{df_col}_idx"] = df[df_col]
            df = df.ordinal_encode(f"{df_col}_idx", values=task_labels, lazy=True)
    return df


def _rename_prob_cols(df: DataFrame, tasks: List[str]) -> DataFrame:
    for ind, task in enumerate(tasks):
        df.rename(f"prob_{ind}", f"prob_{task}")
    return df


def _validate_filter(filter: Union[FilterParams, Dict] = None) -> Dict:
    # Validate the fields provided with pydantic before making request
    return FilterParams(**dict(filter or {})).dict()


def _clean_mltc_df(df: DataFrame) -> DataFrame:
    """In MLTC, don't return the non-task-indexes gold/pred/dep columns"""
    drop_cols = ["gold", "pred", "data_error_potential", "prob"]
    keep_cols = [col for col in df.get_column_names() if col not in drop_cols]
    return df[keep_cols]



from random import randint, sample
from typing import Callable, Tuple
from unittest import mock

import numpy as np
import pandas as pd
import pytest
import vaex

import dataquality
import dataquality as dq
from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger.base_data_logger import DataSet
from dataquality.loggers.data_logger.text_multi_label import TextMultiLabelDataLogger
from dataquality.loggers.model_logger.text_multi_label import TextMultiLabelModelLogger
from dataquality.schemas.split import Split


def test_duplicate_rows(set_test_config, cleanup_after_use) -> None:
    set_test_config(task_type="text_multi_label")
    text_inputs = [
        "what movies star bruce willis",
        "show me films with drew barrymore from the 1980s",
        "what movies starred both al pacino and robert deniro",
        "find me all of the movies that starred harold ramis and bill murray",
        "find me a movie with a quote about baseball in it",
    ]
    gold = [["A", "C", "B"]] * 5
    ids = [0, 1, 2, 3, 4]

    dq.log_data_samples(
        texts=text_inputs, task_labels=gold, split="validation", ids=ids
    )
    dq.log_data_samples(texts=text_inputs, task_labels=gold, split="training", ids=ids)

    with pytest.raises(GalileoException):
        dq.log_data_samples(
            texts=text_inputs, task_labels=gold, split="validation", ids=ids
        )

    dq.log_data_samples(texts=text_inputs, task_labels=gold, split="test", ids=ids)

    with pytest.raises(GalileoException):
        dq.log_data_samples(texts=text_inputs, task_labels=gold, split="test", ids=ids)


def test_duplicate_outputs_rows(set_test_config, cleanup_after_use) -> None:
    set_test_config(task_type="text_multi_label")
    num_samples = 5
    num_tasks = 4
    classes_per_task = 3
    text_inputs = [
        "what movies star bruce willis",
        "show me films with drew barrymore from the 1980s",
        "what movies starred both al pacino and robert deniro",
        "find me all of the movies that starred harold ramis and bill murray",
        "find me a movie with a quote about baseball in it",
    ]
    gold = [["A", "C", "B", "D"]] * 5
    ids = list(range(5))

    dq.log_data_samples(
        texts=text_inputs, task_labels=gold, split="validation", ids=ids
    )
    dq.log_data_samples(texts=text_inputs, task_labels=gold, split="training", ids=ids)

    embs = np.random.rand(num_samples, 100)
    logits = [[np.random.rand(classes_per_task)] * num_tasks] * num_samples
    ids = list(range(5))
    dq.log_model_outputs(embs=embs, logits=logits, ids=ids, split="training", epoch=0)
    dq.log_model_outputs(embs=embs, logits=logits, ids=ids, split="training", epoch=0)

    with pytest.raises(GalileoException) as e:
        dq.get_data_logger().upload()

    assert str(e.value).startswith("It seems as though you do not have unique ids")


def test_log_data_sample(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_multi_label")
    logger = TextMultiLabelDataLogger()
    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dq.log_data_sample(text="sample 1", task_labels="A", id=1, split="training")

        assert logger.texts == ["sample 1"]
        assert logger.labels == [["A"]]
        assert logger.ids == [1]
        assert logger.split == Split.training


@pytest.mark.parametrize(
    "dataset",
    [
        pd.DataFrame(
            {
                "my_text": ["sample1", "sample2", "sample3"],
                "my_labels": [["A", "A", "B"], ["C", "A", "B"], ["B", "C", "B"]],
                "my_id": [1, 2, 3],
            }
        ),
        vaex.from_pandas(
            pd.DataFrame(
                {
                    "my_text": ["sample1", "sample2", "sample3"],
                    "my_labels": [["A", "A", "B"], ["C", "A", "B"], ["B", "C", "B"]],
                    "my_id": [1, 2, 3],
                }
            )
        ),
        [
            {"my_text": "sample1", "my_labels": ["A", "A", "B"], "my_id": 1},
            {"my_text": "sample2", "my_labels": ["C", "A", "B"], "my_id": 2},
            {"my_text": "sample3", "my_labels": ["B", "C", "B"], "my_id": 3},
        ],
    ],
)
def test_log_dataset(
    dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    logger = TextMultiLabelDataLogger()

    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dq.log_dataset(
            dataset, text="my_text", label="my_labels", id="my_id", split="train"
        )

        assert logger.texts == ["sample1", "sample2", "sample3"]
        log_labels = [list(i) for i in logger.labels]
        assert log_labels == [["A", "A", "B"], ["C", "A", "B"], ["B", "C", "B"]]
        assert logger.ids == [1, 2, 3]
        assert logger.split == Split.training


def test_log_dataset_tuple(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    logger = TextMultiLabelDataLogger()
    dataset = [
        ("sample1", ["A", "A", "B"], 1),
        ("sample2", ["C", "A", "B"], 2),
        ("sample3", ["B", "C", "B"], 3),
    ]

    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dq.log_dataset(dataset, text=0, label=1, id=2, split="train")

        assert logger.texts == ["sample1", "sample2", "sample3"]
        log_labels = [list(i) for i in logger.labels]
        assert log_labels == [["A", "A", "B"], ["C", "A", "B"], ["B", "C", "B"]]
        assert logger.ids == [1, 2, 3]
        assert logger.split == Split.training


def test_log_data_binary(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_multi_label")
    logger = TextMultiLabelDataLogger()
    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        tasks = ["A", "B", "C", "D"]
        dq.set_tasks_for_run(tasks, binary=True)
        task_labels = [sample(tasks, k=randint(1, 4)) for _ in range(10)]
        dq.log_data_samples(
            texts=[f"sample {i}" for i in range(10)],
            task_labels=task_labels,
            ids=list(range(10)),
            split="training",
        )

        assert logger.texts == [f"sample {i}" for i in range(10)]
        assert logger.ids == list(range(10))
        for logged_labels, clean_sample_labels in zip(task_labels, logger.labels):
            assert len(clean_sample_labels) == 4
            for task in tasks:
                if task in logged_labels:
                    assert task in clean_sample_labels
                else:
                    assert f"NOT_{task}" in clean_sample_labels
        assert logger.logger_config.observed_num_tasks == 4


def test_log_data_binary_not_setting_binary(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_multi_label")

    tasks = ["A", "B", "C", "D"]
    task_labels = [sample(tasks, k=randint(1, 4)) for _ in range(10)]
    with pytest.raises(AssertionError) as e:
        dq.log_data_samples(
            texts=[f"sample {i}" for i in range(10)],
            task_labels=task_labels,
            ids=list(range(10)),
            split="training",
        )
    # We didn't set binary=true in our tasks so during logging we should get a nice err
    assert str(e.value).startswith(
        "Each training input must have the same number of labels."
    )
    assert "If this is a binary multi label and" in str(e.value)


def test_set_tasks_not_set_binary(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_multi_label")

    tasks = ["A", "B", "C", "D"]
    dq.set_tasks_for_run(tasks)
    # We didn't set binary=True above so validation should fail nicely
    with pytest.raises(ValueError) as e:
        dq.set_labels_for_run(tasks)

    err = e.value.errors()[0]["msg"]
    assert err.startswith("Labels must be a list of lists.")
    assert "If you are running a binary multi-label case," in err


def test_log_model_outputs_binary(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_multi_label")
    tasks = ["A", "B", "C", "D"]
    dq.set_tasks_for_run(tasks, binary=True)

    logger = TextMultiLabelModelLogger(
        embs=np.random.rand(10, 100),  # 10 samples, 100 emb per sample
        logits=np.random.rand(10, 5),  # 10 samples, 5 tasks
        ids=list(range(10)),
        split="training",
        epoch=0,
    )
    logger.logger_config.observed_num_tasks = 5
    logger._log()

    assert not hasattr(logger, "logits")
    assert logger.probs.shape == (10, 5, 2)  # samples X tasks X classes per task
    for sample_probs in logger.probs:
        for task_probs in sample_probs:
            assert np.isclose(np.sum(task_probs), 1.0)


@pytest.mark.parametrize("dims", [(10, 5, 2), (10,)])
def test_log_model_outputs_binary_bad_shapes(
    dims: Tuple, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_multi_label")
    tasks = ["A", "B", "C", "D"]
    dq.set_tasks_for_run(tasks, binary=True)

    logger = TextMultiLabelModelLogger(
        embs=np.random.rand(10, 100),  # 10 samples, 100 emb per sample
        logits=np.random.rand(*dims),
        ids=list(range(10)),
        split="training",
        epoch=0,
    )
    ndim = 5 if len(dims) == 3 else 2
    logger.logger_config.observed_num_tasks = ndim

    with pytest.raises(GalileoException) as e:
        logger._log()

    err = str(e.value)
    if len(dims) == 1:
        assert "Probs/logits must have at least 2 dimensions, they have 1" in err
    else:
        assert (
            f"In binary multi-label, your logits should have 2 dimensions, but they "
            f"currently have {len(dims)}."
        ) in err


@pytest.mark.parametrize("set_labels_first", [True, False])
def test_logged_labels_dont_match_set_labels(
    set_labels_first: bool, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    """An error should be thrown when the set labels dont match the logged labels"""
    set_test_config(task_type="text_multi_label")
    labels = [["A", "NOT_A"], ["B", "NOT_B"], ["C", "NOT_C"]]
    # labels are the index, not the actual labels. No good
    dataset = pd.DataFrame(
        {
            "text": ["sample1", "sample2", "sample3"],
            "label": [[1, 1, 2], [3, 1, 2], [2, 3, 2]],
            "id": [1, 2, 3],
        }
    )
    if set_labels_first:
        dataquality.set_labels_for_run(labels)
        with pytest.raises(AssertionError) as e:
            dataquality.log_dataset(dataset, split="train")
        assert str(e.value).startswith(
            "The input labels you log must be exactly the same"
        )
    else:
        dataquality.log_dataset(dataset, split="train")
        dataquality.get_data_logger().logger_config.observed_num_labels = [2, 2, 2]
        dataquality.set_labels_for_run(labels)
        with pytest.raises(AssertionError) as e:
            dataquality.get_data_logger().validate_labels()
        assert str(e.value).startswith("The labels set for task #")



import json
import os
import warnings
from enum import Enum, unique
from pathlib import Path
from typing import Dict, Optional

import requests
from pydantic import BaseModel
from pydantic.class_validators import validator
from pydantic.types import UUID4
from requests.exceptions import ConnectionError as ReqConnectionError

from dataquality import __version__ as dq_version
from dataquality.exceptions import GalileoException
from dataquality.schemas.task_type import TaskType
from dataquality.utils.helpers import galileo_disabled


class GalileoConfigVars(str, Enum):
    API_URL = "GALILEO_API_URL"
    CONSOLE_URL = "GALILEO_CONSOLE_URL"

    @staticmethod
    def get_config_mapping() -> Dict[str, Optional[str]]:
        return {i.name.lower(): os.environ.get(i.value) for i in GalileoConfigVars}

    @staticmethod
    def get_available_config_attrs() -> Dict[str, str]:
        return {
            i.name.lower(): os.environ.get(i.value, "")
            for i in GalileoConfigVars
            if os.environ.get(i.value)
        }

    @staticmethod
    def auto_init_vars_available() -> bool:
        return bool(os.getenv("GALILEO_API_URL"))


class ConfigData(str, Enum):
    DEFAULT_GALILEO_CONFIG_DIR = f"{os.environ.get('HOME', str(Path.home()))}/.galileo"
    DEFAULT_GALILEO_CONFIG_FILE = f"{DEFAULT_GALILEO_CONFIG_DIR}/config.json"


@unique
class AuthMethod(str, Enum):
    email = "email"


class Config(BaseModel):
    api_url: str
    auth_method: AuthMethod = AuthMethod.email
    token: Optional[str] = None
    current_user: Optional[str] = None
    current_project_id: Optional[UUID4] = None
    current_run_id: Optional[UUID4] = None
    task_type: Optional[TaskType] = None

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True

    def update_file_config(self) -> None:
        config_json = self.dict()

        with open(ConfigData.DEFAULT_GALILEO_CONFIG_FILE.value, "w+") as f:
            f.write(json.dumps(config_json, default=str))

    @validator("api_url", pre=True, always=True, allow_reuse=True)
    def add_scheme(cls, v: str) -> str:
        if not v.startswith("http"):
            # api url needs the scheme
            v = f"http://{v}"
        return v


def url_is_localhost(url: str) -> bool:
    return any(["localhost" in url, "127.0.0.1" in url])


def set_platform_urls(console_url_str: str) -> None:
    if url_is_localhost(console_url_str):
        os.environ[GalileoConfigVars.API_URL] = "http://localhost:8088"
    else:
        api_url = console_url_str.replace("console.", "api.").rstrip("/")
        _validate_api_url(console_url_str, api_url)
        os.environ[GalileoConfigVars.API_URL] = api_url


def _validate_api_url(console_url: str, api_url: str) -> None:
    """Ensures the api url is a valid one"""
    err_detail = (
        f"The provided console URL {console_url} is invalid or is not currently "
        "available. If you are sure this is correct, reach out to your admin.\n\n"
        "To change your console url, run: \n`os.environ['GALILEO_CONSOLE_URL']='URL'` "
        "and then \n`dq.configure()`"
        "\n\nDetail: {err}"
    )
    url = f"http://{api_url}" if not api_url.startswith("http") else api_url
    try:
        r = requests.get(f"{url}/healthcheck")
        if not r.ok:
            raise GalileoException(err_detail.format(err=r.text)) from None
    except (ReqConnectionError, ConnectionError) as e:
        raise GalileoException(err_detail.format(err=str(e))) from None


def _check_console_url() -> None:
    """Checks for user setting of GALILEO_CONSOLE_URL instead of

    GALILEO_API_URL. If set, this will automatically set
    platform urls (GALILEO_API_URL) for auto_init
    """
    console_url = os.getenv(GalileoConfigVars.CONSOLE_URL)
    if console_url:
        if "console." not in console_url and not url_is_localhost(console_url):
            warnings.warn(
                f"It seems your GALILEO_CONSOLE_URL ({console_url}) is invalid. "
                f"Your console URL should have 'console.' in the url. Ignoring"
            )
        else:
            set_platform_urls(console_url_str=console_url)


def set_config() -> Config:
    if galileo_disabled():
        return Config(api_url="")
    _check_console_url()
    if not os.path.isdir(ConfigData.DEFAULT_GALILEO_CONFIG_DIR.value):
        os.makedirs(ConfigData.DEFAULT_GALILEO_CONFIG_DIR.value, exist_ok=True)
    if os.path.exists(ConfigData.DEFAULT_GALILEO_CONFIG_FILE.value):
        with open(ConfigData.DEFAULT_GALILEO_CONFIG_FILE.value) as f:
            try:
                config_vars: Dict[str, str] = json.load(f)
            # If there's an issue reading the config file for any reason, quit and
            # start fresh
            except Exception as e:
                warnings.warn(
                    f"We had an issue reading your config file ({type(e)}). "
                    f"Recreating your file from scratch."
                )
                return reset_config()
        # If the user updated any config vars via env, grab those updates
        new_config_attrs = GalileoConfigVars.get_available_config_attrs()
        config_vars.update(**new_config_attrs)
        config = Config(**config_vars)

    elif GalileoConfigVars.auto_init_vars_available():
        galileo_vars = GalileoConfigVars.get_config_mapping()
        config = Config(**galileo_vars)

    else:
        print(f"Welcome to Galileo {dq_version}!")
        print(
            "To skip this prompt in the future, set the following environment "
            "variable: GALILEO_CONSOLE_URL"
        )
        console_url = input("🔭 Enter the url of your Galileo console\n")
        set_platform_urls(console_url_str=console_url)
        galileo_vars = GalileoConfigVars.get_config_mapping()
        config = Config(**galileo_vars)
    config.update_file_config()
    return config


def reset_config() -> Config:
    """Wipe the config file and reconfigure"""
    if os.path.isfile(ConfigData.DEFAULT_GALILEO_CONFIG_FILE.value):
        os.remove(ConfigData.DEFAULT_GALILEO_CONFIG_FILE.value)
    return set_config()


config = set_config()



from tempfile import NamedTemporaryFile

import requests
from vaex.dataframe import DataFrame

from dataquality.core.auth import api_client
from dataquality.utils.file import get_file_extension


class ObjectStore:
    ROOT_BUCKET_NAME = "galileo-project-runs"
    DOWNLOAD_CHUNK_SIZE_MB = 256

    def create_project_run_object(
        self,
        object_name: str,
        file_path: str,
        content_type: str = "application/octet-stream",
    ) -> None:
        url = api_client.get_presigned_url(
            project_id=object_name.split("/")[0],
            method="put",
            bucket_name=self.ROOT_BUCKET_NAME,
            object_name=object_name,
        )
        self._upload_file_from_local(
            url=url, file_path=file_path, content_type=content_type
        )

    def _upload_file_from_local(
        self, url: str, file_path: str, content_type: str = "application/octet-stream"
    ) -> None:
        """_upload_file_from_local

        Args:
            url (str): The url to request.
            file_path (str): Where data is stored on the local file system.
            content_type (str): The content type of the upload request.

        Returns:
            None
        """
        requests.put(
            url=url, data=open(file_path, "rb"), headers={"content-type": content_type}
        )

    def create_project_run_object_from_df(
        self, df: DataFrame, object_name: str
    ) -> None:
        """Uploads a Vaex dataframe at the specified object_name location"""
        ext = get_file_extension(object_name)
        with NamedTemporaryFile(suffix=ext) as f:
            df.export(f.name)
            self.create_project_run_object(
                object_name=object_name,
                file_path=f.name,
            )

    def download_file(self, object_name: str, file_path: str) -> str:
        """download_file

        Args:
            object_name (str): The object name.
            file_path (str): Where to write the object data locally.

        Returns:
            str: The local file where the object name was written.
        """
        url = api_client.get_presigned_url(
            project_id=object_name.split("/")[0],
            method="get",
            bucket_name=self.ROOT_BUCKET_NAME,
            object_name=object_name,
        )
        return self._local_download_from_url(url=url, file_path=file_path)

    def _local_download_from_url(self, url: str, file_path: str) -> str:
        """_local_download_from_url

        Args:
            url (str): The url to request.
            file_path (str): The path to where data was streamed on the requester's
             local filesystem.

        Returns:
            str: The path to where data was streamed on the requester's local
             filesystem.
        """
        with requests.get(url, stream=True) as remote_file:
            remote_file.raise_for_status()
            with open(file_path, "wb") as local_file:
                for chunk in remote_file.iter_content(
                    chunk_size=1024 * self.DOWNLOAD_CHUNK_SIZE_MB
                ):
                    local_file.write(chunk)
        return file_path



import logging
import os
from typing import Any, Optional, Tuple

from pydantic import UUID4

from dataquality.clients.api import ApiClient
from dataquality.clients.objectstore import ObjectStore
from dataquality.core._config import ConfigData, config
from dataquality.utils.helpers import check_noop

DQ_LOG_FILE_HOME = f"{ConfigData.DEFAULT_GALILEO_CONFIG_DIR}/out"
DQ_LOG_FILE = "out.log"


class CustomSplitAdapter(logging.LoggerAdapter):
    """
    This adapter appends the split to the message, if found. Otherwise, "None"

    Adapted from https://docs.python.org/3/howto/logging-cookbook.html (CustomAdapter)
    """

    def process(self, msg: str, kwargs: Any) -> Tuple[str, Any]:
        split = kwargs.pop("split", self.extra["split"])
        epoch = kwargs.pop("epoch", self.extra["epoch"])
        if epoch is not None:
            return "[%s] [epoch:%s]: %s" % (split, str(epoch), msg), kwargs
        return "[%s]: %s" % (split, msg), kwargs


log_formatter = logging.Formatter(
    "%(asctime)s [%(levelname)s] [%(threadName)s] %(message)s"
)


def get_dq_logger() -> CustomSplitAdapter:
    """Returns the dq logger for the current run_id"""
    logger = logging.getLogger(str(config.current_run_id))
    logger.setLevel(os.environ.get("GALILEO_LOG_LEVEL", "INFO").upper())
    # Avoid adding multiple handlers if one already exists
    if not logger.handlers:
        handler = logging.FileHandler(dq_log_file_path())
        handler.setFormatter(log_formatter)
        logger.addHandler(handler)
    adapter = CustomSplitAdapter(logger, {"split": None, "epoch": None})
    return adapter


def dq_log_file_path(run_id: Optional[UUID4] = None) -> str:
    rid = run_id or config.current_run_id
    return f"{DQ_LOG_FILE_HOME}/{rid}/{DQ_LOG_FILE}"


def remove_dq_log_file(run_id: Optional[UUID4] = None) -> None:
    file_path = dq_log_file_path(run_id)
    if os.path.isfile(file_path):
        os.remove(file_path)


def dq_log_object_name(project_id: UUID4, run_id: UUID4) -> str:
    """Returns the minio/s3 object name"""
    return f"{project_id}/{run_id}/out/{DQ_LOG_FILE}"


@check_noop
def upload_dq_log_file() -> None:
    # For typing
    assert config.current_project_id and config.current_run_id
    obj_store = ObjectStore()
    obj_name = dq_log_object_name(config.current_project_id, config.current_run_id)
    file_path = dq_log_file_path()
    if os.path.isfile(file_path):
        obj_store.create_project_run_object(
            object_name=obj_name, file_path=file_path, content_type="text/plain"
        )
        remove_dq_log_file()


@check_noop
def get_dq_log_file(
    project_name: Optional[str] = None, run_name: Optional[str] = None
) -> str:
    pid, rid = ApiClient()._get_project_run_id(
        project_name=project_name, run_name=run_name
    )
    log_object = dq_log_object_name(pid, rid)
    log_file_path = dq_log_file_path(rid)
    obj_store = ObjectStore()
    obj_store.download_file(object_name=log_object, file_path=log_file_path)
    print(f"Your logfile has been written to {log_file_path}")
    return log_file_path



from random import random

import numpy as np
import pandas as pd
import pytorch_lightning as pl
import torch
from sklearn.datasets import fetch_20newsgroups

import dataquality

NUM_RECORDS = 23


dataquality.config.task_type = "text_classification"


def introduce_label_errors(
    df: pd.DataFrame, column: str, shuffle_percent: int
) -> pd.DataFrame:
    arr = df[column].values
    shuffle = np.random.choice(
        np.arange(arr.shape[0]),
        round(arr.shape[0] * shuffle_percent / 100),
        replace=False,
    )
    arr[np.sort(shuffle)] = arr[shuffle]
    df[column] = arr
    return df


class NewsgroupDataset(torch.utils.data.Dataset):
    def __init__(self, split: str) -> None:
        newsgroups = fetch_20newsgroups(
            subset="train" if split == "training" else "test",
            remove=("headers", "footers", "quotes"),
        )

        self.dataset = pd.DataFrame()
        self.dataset["text"] = newsgroups.data
        self.dataset["label"] = newsgroups.target
        self.dataset = self.dataset[:NUM_RECORDS]
        self.glogger = dataquality.get_data_logger()

        # Shuffle some percentage of the training dataset
        # to force create mislabeled samples
        if split == "training":
            self.dataset = introduce_label_errors(self.dataset, "label", 11)

        #
        # 🔭 Logging Inputs with Galileo!
        #
        self.glogger.texts = self.dataset["text"]
        self.glogger.labels = self.dataset["label"]
        self.glogger.ids = list(range(len(self.dataset)))

    def __getitem__(self, idx):
        x = torch.tensor(np.random.randint(0, 100, size=50))
        attention_mask = x
        y = self.dataset["label"][idx]
        return idx, x, attention_mask, y

    def __len__(self):
        return len(self.dataset)


class TorchDistilBERTTorch(pl.LightningModule):
    def __init__(self):
        super().__init__()

    def forward(self, x, attention_mask, x_idxs):
        # Fake model building for less memory usage
        logits = [[random() for _ in range(5)] for _ in range(NUM_RECORDS)]
        embs = [[random() for _ in range(10)] for _ in range(NUM_RECORDS)]

        # Logging with Galileo!
        self.glogger = dataquality.get_model_logger()(
            embs=embs, logits=logits, ids=x_idxs
        )

        return 0


class LightningDistilBERT(pl.LightningModule):
    def __init__(self):
        super().__init__()

    def forward(self, x, attention_mask, x_idxs):
        # Fake model building for less memory usage
        logits = [[random() for _ in range(5)] for _ in range(NUM_RECORDS)]
        embs = [[random() for _ in range(10)] for _ in range(NUM_RECORDS)]

        # Logging with Galileo!
        self.glogger = dataquality.get_model_logger()(
            embs=embs, logits=logits, ids=x_idxs
        )

        return 0

    def training_step(self, batch, batch_idx):
        """Model training step."""
        x_idxs, x, attention_mask, y = batch
        self(
            x=x,
            attention_mask=attention_mask,
            x_idxs=x_idxs,
        )

    def validation_step(self, batch, batch_idx):
        """Model validation step."""
        x_idxs, x, attention_mask, y = batch
        self(
            x=x,
            attention_mask=attention_mask,
            x_idxs=x_idxs,
        )

    def test_step(self, batch, batch_idx):
        """Model test step."""
        x_idxs, x, attention_mask, y = batch
        self(
            x=x,
            attention_mask=attention_mask,
            x_idxs=x_idxs,
        )

    def configure_optimizers(self):
        """Model optimizers."""


model = LightningDistilBERT()
torch_model = TorchDistilBERTTorch()



from typing import List, Optional, Set

import numpy as np
from pydantic import validator

from dataquality.loggers.logger_config.base_logger_config import BaseLoggerConfig


class TextClassificationLoggerConfig(BaseLoggerConfig):
    labels: Optional[List[str]] = None
    observed_num_labels: int = 0
    observed_labels: Set[str] = set()

    class Config:
        validate_assignment = True

    @validator("labels", always=True, pre=True, allow_reuse=True)
    def clean_labels(cls, labels: List[str]) -> List[str]:
        if labels is None:
            return labels
        if isinstance(labels, np.ndarray):
            labels = labels.tolist()
        if labels is not None:
            assert isinstance(labels, List), "Labels must be a list"
        if labels and len(labels) == 1:
            labels = [f"NOT_{labels[0]}", labels[0]]
        return [str(i) for i in labels]


text_classification_logger_config = TextClassificationLoggerConfig()



import time
from typing import Callable, Dict, List
from unittest import mock
from unittest.mock import MagicMock, patch
from uuid import uuid4

import pytest

import dataquality.clients.api
from dataquality import config
from dataquality.clients.api import ApiClient
from dataquality.exceptions import GalileoException
from dataquality.schemas.task_type import TaskType
from tests.utils.mock_request import (
    EXISTING_PROJECT,
    EXISTING_RUN,
    FAKE_NEW_RUN,
    MockResponse,
    mocked_delete_project_not_found,
    mocked_delete_project_run,
    mocked_get_project_run,
    mocked_missing_project_name,
)

api_client = ApiClient()


@mock.patch("requests.get", side_effect=mocked_get_project_run)
@mock.patch("requests.delete", side_effect=mocked_delete_project_run)
def test_delete_project(
    mock_delete_run: MagicMock, mock_get_run: MagicMock, set_test_config: Callable
) -> None:
    """Base case: Tests creating a new project and run"""
    api_client.delete_project(uuid4())


@mock.patch("requests.get", side_effect=mocked_delete_project_not_found)
@mock.patch("requests.delete", side_effect=mocked_delete_project_not_found)
def test_delete_project_not_found(
    mock_delete_run: MagicMock,
    mock_get_run: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    with pytest.raises(GalileoException):
        api_client.delete_project(uuid4())


@mock.patch("requests.get", side_effect=mocked_get_project_run)
@mock.patch("requests.delete", side_effect=mocked_delete_project_run)
def test_delete_project_by_name(
    mock_delete_run: MagicMock,
    mock_get_run: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    api_client.delete_project_by_name(EXISTING_PROJECT)


@mock.patch("requests.get", side_effect=mocked_missing_project_name)
def test_delete_project_by_name_not_found(
    mock_get_run: MagicMock, set_test_config: Callable
) -> None:
    """Base case: Tests creating a new project and run"""
    with pytest.raises(GalileoException):
        api_client.delete_project_by_name("some_proj")


@mock.patch("requests.get", side_effect=mocked_get_project_run)
@mock.patch("requests.delete", side_effect=mocked_delete_project_run)
def test_delete_run(
    mock_delete_run: MagicMock,
    mock_get_run: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    api_client.delete_run(uuid4(), uuid4())


@mock.patch("requests.get", side_effect=mocked_get_project_run)
@mock.patch("requests.delete", side_effect=mocked_delete_project_not_found)
def test_delete_run_missing_run(
    mock_delete_run: MagicMock,
    mock_get_run: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    with pytest.raises(GalileoException):
        api_client.delete_run(uuid4(), uuid4())


@mock.patch("requests.get", side_effect=mocked_get_project_run)
@mock.patch("requests.delete", side_effect=mocked_delete_project_run)
def test_delete_run_by_name(
    mock_delete_run: MagicMock,
    mock_get_run: MagicMock,
    set_test_config: Callable,
) -> None:
    """Base case: Tests creating a new project and run"""
    api_client.delete_run_by_name(EXISTING_PROJECT, EXISTING_RUN)


@mock.patch("requests.get")
def test_delete_run_by_name_missing_run(
    mock_get_run: MagicMock, set_test_config: Callable
) -> None:
    """tests deleting a project/run with a run name that doesn't exist"""
    mock_get_run.side_effect = [
        MockResponse(json_data=[{"id": "uuid"}], status_code=200),
        MockResponse(json_data=[], status_code=200),
    ]
    with pytest.raises(GalileoException) as e:
        api_client.delete_run_by_name("some_proj", "some_run")

    assert str(e.value) == "No project/run found with name some_proj/some_run"


@pytest.mark.parametrize(
    "api_url",
    ["http://localhost:8088", "http://127.0.0.1:8088"],
)
def test_get_run_status_localhost_fails(
    set_test_config: Callable, api_url: str
) -> None:
    """Raises error if api url is set to localhost"""
    set_test_config(default_api_url=api_url)
    with pytest.raises(GalileoException):
        api_client.get_run_status()


@pytest.mark.parametrize(
    "project_name, run_name",
    [("carrots", None), (None, "Rhubarb")],
)
def test_get_run_status_project_or_run_fails(project_name: str, run_name: str) -> None:
    """Raises error when exactly one of project_name and run_name is passed in"""
    with pytest.raises(GalileoException):
        api_client.get_run_status(project_name, run_name)


@mock.patch.object(ApiClient, "make_request")
def test_get_run_status_project_default_current_run(
    mock_make_request: MagicMock,
    set_test_config: Callable,
    statuses_response: Dict[str, List],
) -> None:
    """Happy path: Empty args return status for current run"""
    mock_make_request.return_value = statuses_response
    project_id = uuid4()
    run_id = uuid4()
    set_test_config(
        current_project_id=project_id,
        current_run_id=run_id,
        api_url="https://api.fake.com",
    )
    status = api_client.get_run_status()
    assert status["status"] == "finished"
    mock_make_request.assert_called_once_with(
        "get", f"https://api.fake.com/projects/{project_id}/runs/{run_id}/jobs/status"
    )


@mock.patch.object(ApiClient, "get_project_run_by_name")
@mock.patch.object(ApiClient, "make_request")
def test_get_run_status_project_specified_run(
    mock_make_request: MagicMock,
    mock_get_project_run_by_name: MagicMock,
    set_test_config: Callable,
    statuses_response: Dict[str, List],
) -> None:
    # TODO
    """Happy path: Specifying a project, run different from current run"""
    mock_make_request.return_value = statuses_response
    config_project_id = uuid4()
    config_run_id = uuid4()
    project_id = uuid4()
    run_id = uuid4()
    mock_get_project_run_by_name.return_value = {"project_id": project_id, "id": run_id}

    set_test_config(
        current_project_id=project_id,
        current_run_id=run_id,
        api_url="https://api.fake.com",
    )
    status = api_client.get_run_status("Carrot", "Rhubarb")
    assert status["status"] == "finished"
    assert config_project_id != project_id
    assert config_run_id != run_id
    mock_make_request.assert_called_once_with(
        "get", f"https://api.fake.com/projects/{project_id}/runs/{run_id}/jobs/status"
    )


@mock.patch.object(ApiClient, "get_project_run_by_name", return_value={})
def test_get_run_status_project_nonexistent_run_fails(
    mock_get_project_run_by_name: MagicMock, set_test_config: Callable
) -> None:
    """Specifying a project, run that doesn't exist raises error"""
    set_test_config(api_url="https://api.fake.com")
    with pytest.raises(GalileoException):
        api_client.get_run_status("Carrot", "Fake-Run")


@mock.patch.object(
    ApiClient,
    "get_run_status",
    side_effect=[{"status": "started"}, {"status": "finished"}],
)
def test_wait_for_run_started_finished(mock_get_run_status: MagicMock) -> None:
    """Happy path: Returns after transitioning from started to finished"""
    t_start = time.time()
    api_client.wait_for_run("some_proj", "some_run")
    assert mock_get_run_status.call_count == 2
    assert time.time() - t_start > 2


@mock.patch.object(ApiClient, "get_run_status", return_value={"status": "errored"})
def test_wait_for_run_errored(mock_get_run_status: MagicMock) -> None:
    """Waiting for run with errored status raises error"""
    with pytest.raises(GalileoException):
        api_client.wait_for_run("some_proj", "some_run")


@mock.patch.object(ApiClient, "get_run_status", return_value={"status": "unknown"})
def test_wait_for_run_unknown(mock_get_run_status: MagicMock) -> None:
    """Waiting for run with unknown status raises error"""
    with pytest.raises(GalileoException):
        api_client.wait_for_run("some_proj", "some_run")


@patch.object(
    ApiClient,
    "delete_run",
)
@patch.object(
    ApiClient,
    "create_run",
    return_value={"name": "my_run", "id": FAKE_NEW_RUN},
)
@patch.object(
    ApiClient,
    "get_project",
    return_value={"name": "my_project"},
)
@patch.object(
    ApiClient,
    "get_project_run",
    return_value={"name": "my_run", "task_type": 0},
)
def test_reset(
    mock_get_run: MagicMock,
    mock_get_project: MagicMock,
    mock_create: MagicMock,
    mock_delete: MagicMock,
    set_test_config: Callable,
) -> None:
    """
    Tests that reset run changes the run ID saved and updates the config
    """
    old_pid = config.current_project_id
    assert config.current_run_id != FAKE_NEW_RUN
    old_rid = config.current_run_id
    api_client.reset_run(old_pid, old_rid)

    mock_get_run.assert_called_once_with(old_pid, old_rid)
    mock_get_project.assert_called_once_with(old_pid)
    mock_create.assert_called_once_with(
        "my_project", "my_run", TaskType.text_classification
    )
    mock_delete.assert_called_once_with(old_pid, old_rid)

    assert config.current_run_id == FAKE_NEW_RUN


@patch.object(
    ApiClient,
    "get_run_status",
    side_effect=[{}, {"status": "finished", "timestamp": 1}],
)
def test_get_run_status_no_status(
    mock_get_status: MagicMock, set_test_config: Callable
) -> None:
    """Asserts that wait_for_run with an empty status doens't crash"""
    api_client.wait_for_run()
    assert mock_get_status.call_count == 2


@patch.object(dataquality.clients.api.requests, "post")
@patch.object(dataquality.clients.api.ApiClient, "get_task_type")
@patch.object(dataquality.clients.api.ApiClient, "_get_project_run_id")
def test_export_run_no_data(
    mock_get_run: MagicMock, mock_get_task_type: MagicMock, mock_post: MagicMock
) -> None:
    mock_get_run.return_value = uuid4(), uuid4()
    mock_get_task_type.return_value = TaskType.text_classification
    # In export_run we use requests.post as a context manager (with requests.post(...))
    # so we need to mock the `__enter__` return value
    mock_post.return_value.__enter__.return_value = MockResponse(
        status_code=200, json_data={}, headers={"Galileo-No-Data": "true"}
    )
    with pytest.raises(GalileoException) as e:
        api_client.export_run("project", "run", "training", "file.csv")
    assert str(e.value).startswith("It seems there is no data for this request")



from typing import Callable

import dataquality
from dataquality.loggers import BaseGalileoLogger
from dataquality.loggers.data_logger import BaseGalileoDataLogger
from dataquality.loggers.model_logger import BaseGalileoModelLogger


def test_attribute_subsets() -> None:
    """All potential logging fields used by all subclass loggers should be encapsulated

    Any new logger that is created has a set of attributes that it expects from users.
    The `BaseLoggerAttributes` from the BaseGalileoLogger should be the superset of
    all child loggers.
    """
    all_attrs = set(BaseGalileoLogger.get_valid_attributes())
    sub_data_loggers = BaseGalileoDataLogger.__subclasses__()
    data_logger_attrs = set(
        [j for i in sub_data_loggers for j in i.get_valid_attributes()]
    )
    sub_model_loggers = BaseGalileoModelLogger.__subclasses__()
    model_logger_attrs = set(
        [j for i in sub_model_loggers for j in i.get_valid_attributes()]
    )
    all_sub_attrs = data_logger_attrs.union(model_logger_attrs)
    assert all_attrs.issuperset(
        all_sub_attrs
    ), f"Missing attrs: {all_sub_attrs - all_attrs}"


def test_int_labels(set_test_config: Callable) -> None:
    dataquality.set_labels_for_run(labels=[1, 2, 3, 4, 5])  # type: ignore
    assert dataquality.get_data_logger().logger_config.labels == [
        "1",
        "2",
        "3",
        "4",
        "5",
    ]



import requests
from setuptools.version import pkg_resources

from dataquality import __version__ as dq_client_version
from dataquality.core._config import config
from dataquality.exceptions import GalileoException
from dataquality.schemas.route import Route


def _version_check() -> None:
    """_version_check.

    Asserts that the dataquality python client and the api have
    matching major versions.

    https://semver.org/#summary.

    Returns:
        None
    """
    client_semver = _parse_version(_get_client_version())
    server_semver = _parse_version(_get_api_version())
    try:
        assert _major_version(client_semver) == _major_version(server_semver)
    except AssertionError:
        raise GalileoException(
            "Major mismatch between client, "
            f"{client_semver}, and server {server_semver}."
        )


def _major_version(v: pkg_resources.extern.packaging.version.Version) -> str:
    if hasattr(v, "major"):
        return str(v.major)
    else:
        return str(v.base_version).split(".")[0]


def _parse_version(version: str) -> pkg_resources.extern.packaging.version.Version:
    return pkg_resources.parse_version(version)


def _get_client_version() -> str:
    return dq_client_version


def _get_api_version() -> str:
    response = requests.get(f"{config.api_url}/{Route.healthcheck}")
    response_body = response.json()
    return response_body["api_version"]



import os
from typing import Callable
from unittest import mock
from unittest.mock import MagicMock

import pytest

import dataquality
from dataquality.core.auth import GALILEO_AUTH_METHOD
from dataquality.exceptions import GalileoException
from tests.utils.mock_request import mocked_failed_login_requests, mocked_login_requests

config = dataquality.config


@mock.patch("requests.post", side_effect=mocked_login_requests)
@mock.patch("requests.get", side_effect=mocked_login_requests)
def test_good_login(
    mock_get_current_user: MagicMock, mock_login: MagicMock, set_test_config: Callable
) -> None:
    os.environ[GALILEO_AUTH_METHOD] = "email"
    os.environ["GALILEO_USERNAME"] = "user"
    os.environ["GALILEO_PASSWORD"] = "password"
    set_test_config(token="mytoken")
    dataquality.login()


@mock.patch("requests.post", side_effect=mocked_failed_login_requests)
def test_bad_login(mock_post: MagicMock, set_test_config: Callable) -> None:
    set_test_config(token=None)
    os.environ[GALILEO_AUTH_METHOD] = "email"
    os.environ["GALILEO_USERNAME"] = "user"
    os.environ["GALILEO_PASSWORD"] = "password"
    with pytest.raises(GalileoException) as e:
        dataquality.login()
    assert e.value.args[0] == "Issue authenticating: Incorrect login credentials."



import os
from functools import partial
from random import random
from typing import List, Optional

import numpy as np
import pandas as pd
import vaex
from sklearn.datasets import fetch_20newsgroups
from tqdm import tqdm

import dataquality
from dataquality.schemas.split import Split
from tests.conftest import LOCATION, SPLITS, SUBDIRS, TEST_PATH

NUM_RECORDS = 50
NUM_LOGS = 30
MULTI_LABEL_NUM_TASKS = 4


def validate_uploaded_data(
    expected_num_records: int = None,
    meta_cols: Optional[List] = None,
    multi_label=False,
) -> None:
    """
    Checks for testing
    """
    expected_num_records = expected_num_records or NUM_RECORDS * NUM_LOGS
    meta_cols = meta_cols or {}
    for split in SPLITS:
        # Output data
        split_output_data = {}
        for subdir in SUBDIRS:
            # epoch = 0 for general testing
            try:
                file_path = f"{TEST_PATH}/{split}/0/{subdir}/{subdir}.hdf5"
                data = vaex.open(file_path)
            except FileNotFoundError:
                # Handle autolog test
                file_path = f"{TEST_PATH}/{split}/1/{subdir}/{subdir}.hdf5"
                data = vaex.open(file_path)

            prob_cols = data.get_column_names(regex="prob*")
            for c in data.get_column_names():
                if c in prob_cols + ["emb"]:
                    assert not np.isnan(data[c].values).any()
                else:
                    vals = data[c].values
                    assert all([i is not None and i != "nan" for i in vals])
            split_output_data[subdir] = data

        data = split_output_data["data"]
        emb = split_output_data["emb"]
        prob = split_output_data["prob"]

        for c in meta_cols:
            assert c in data.get_column_names()
        assert list(emb.get_column_names()) == ["id", "emb"]

        assert "data_schema_version" in data.columns
        assert len(data) == len(emb) == len(prob) == expected_num_records
        assert (
            sorted(data["id"].unique())
            == sorted(emb["id"].unique())
            == sorted(prob["id"].unique())
        )
        # Since we log logits, the probs should all sum to 1
        for col in prob.get_column_names(regex="prob*"):
            probs = prob[col].values
            assert np.allclose(1, np.sum(probs, axis=-1))
        if multi_label:
            for c in prob.get_column_names():
                assert c == "id" or c.startswith("prob_") or c.startswith("gold_")
                assert c != "prob" and c != "gold"
            assert (
                len(prob.get_column_names(regex="prob*"))
                == len(prob.get_column_names(regex="gold*"))
                == len(data.get_column_names(regex="pred*"))
                == MULTI_LABEL_NUM_TASKS
            )


def validate_cleanup_data():
    """
    Checks for testing
    """
    for split in SPLITS:
        # Ensure files were cleaned up
        assert not os.path.isdir(f"{LOCATION}/{split}")


def _log_text_classification_data(
    num_records=NUM_RECORDS,
    num_logs=NUM_LOGS,
    unique_ids=True,
    num_embs=20,
    meta=None,
    multi_label=False,
) -> None:
    """
    Logs mock data for text classification/multi-label to disk
    """
    meta = meta or {}
    # Log train/test data
    num_labels_in_task = np.random.randint(low=1, high=10, size=MULTI_LABEL_NUM_TASKS)
    for split in [Split.test, Split.training]:
        newsgroups_train = fetch_20newsgroups(
            subset="train" if split == Split.training else split.value,
            remove=("headers", "footers", "quotes"),
        )
        assert num_records * num_logs <= len(
            newsgroups_train.data
        ), f"num_records*num_logs must be less than {len(newsgroups_train.data)} "
        dataset = pd.DataFrame()
        dataset["text"] = newsgroups_train.data
        dataset["label"] = newsgroups_train.target
        dataset = dataset[: num_records * num_logs]

        if multi_label:
            labels = []
            for i in range(len(dataset)):
                task_labels = [np.random.randint(i) for i in num_labels_in_task]
                labels.append(task_labels)
            dataquality.set_tasks_for_run(
                [str(i) for i in range(MULTI_LABEL_NUM_TASKS)]
            )
        else:
            labels = dataset["label"]
            dataquality.set_labels_for_run([str(i) for i in range(len(set(labels)))])
        dataquality.set_split(split)
        func = partial(
            dataquality.log_data_samples,
            texts=dataset["text"],
            ids=list(range(len(dataset))),
            meta=meta,
        )
        if multi_label:
            func(task_labels=labels)
        else:
            func(labels=labels)

    if multi_label:
        run_labels = [[str(i) for i in range(tl)] for tl in num_labels_in_task]
        dataquality.set_labels_for_run(run_labels)
    for split in [Split.training, Split.test]:
        for ln in tqdm(range(num_logs)):
            embs = [[random() for _ in range(num_embs)] for _ in range(num_records)]
            if multi_label:
                logits = []
                for i in range(num_records):
                    logits_per_task = []
                    for num_labels in num_labels_in_task:
                        logits_per_task.append(np.random.rand(num_labels))
                    logits.append(logits_per_task)
            else:
                logits = np.random.rand(num_records, len(set(labels)))
            epoch = 0

            # Need unique ids
            if unique_ids:
                r = range(ln * num_records, (ln + 1) * num_records)
                ids = list(r)
            else:
                ids = list(range(num_records))

            dataquality.set_epoch(epoch)
            dataquality.set_split(split)
            dataquality.log_model_outputs(embs=embs, logits=logits, ids=ids)



import os
import re
import shutil
import warnings
from typing import Dict, Optional

from pydantic.types import UUID4

import dataquality
from dataquality.clients.api import ApiClient
from dataquality.core._config import config
from dataquality.core.auth import login
from dataquality.exceptions import GalileoException
from dataquality.loggers import BaseGalileoLogger
from dataquality.schemas.task_type import TaskType
from dataquality.utils.dq_logger import DQ_LOG_FILE_HOME
from dataquality.utils.helpers import check_noop
from dataquality.utils.name import random_name

api_client = ApiClient()


class _Init:
    def get_project_by_name_for_user(self, project_name: str) -> Dict:
        return api_client.get_project_by_name(project_name)

    def get_project_run_by_name_for_user(
        self, project_name: str, run_name: str
    ) -> Dict:
        return api_client.get_project_run_by_name(project_name, run_name)

    def _initialize_new_project(
        self, project_name: str, is_public: bool = True
    ) -> Dict:
        visability = "public" if is_public else "private"
        print(f"✨ Initializing {visability} project {project_name}")
        try:
            return api_client.create_project(
                project_name=project_name, is_public=is_public
            )
        except GalileoException as e:
            if "A project with this name already exists" in str(e):
                return api_client.get_project_by_name(project_name)
            else:
                raise e

    def _initialize_run_for_project(
        self, project_name: str, run_name: str, task_type: TaskType
    ) -> Dict:
        print(f"🏃‍♂️ Starting run {run_name}")
        return api_client.create_run(project_name, run_name, task_type)

    def create_log_file_dir(
        self, project_id: UUID4, run_id: UUID4, overwrite_local: bool
    ) -> None:
        write_output_dir = f"{BaseGalileoLogger.LOG_FILE_DIR}/{project_id}/{run_id}"
        stdout_dir = f"{DQ_LOG_FILE_HOME}/{run_id}"
        for out_dir in [write_output_dir, stdout_dir]:
            if overwrite_local and os.path.exists(out_dir):
                shutil.rmtree(out_dir)
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)

    def validate_name(self, name: Optional[str]) -> None:
        """Validates project/run name ensuring only letters, numbers, space, - and _"""
        if not name:
            return
        badchars = re.findall(r"[^\w -]+", name)
        if badchars:
            raise GalileoException(
                "Only letters, numbers, whitespace, - and _ are allowed in a project "
                f"or run name. Remove the following characters: {badchars}"
            )


@check_noop
def init(
    task_type: str,
    project_name: Optional[str] = None,
    run_name: Optional[str] = None,
    is_public: bool = True,
    overwrite_local: bool = True,
) -> None:
    """
    Start a run

    Initialize a new run and new project, initialize a new run in an existing project,
    or reinitialize an existing run in an existing project.

    Optionally provide project and run names to create a new project/run or restart
    existing ones.

    :param task_type: The task type for modeling. This must be one of the valid
    `dataquality.schemas.task_type.TaskType` options
    :param project_name: The project name. If not passed in, a random one will be
    generated. If provided, and the project does not exist, it will be created. If it
    does exist, it will be set.
    :param run_name: The run name. If not passed in, a random one will be
    generated. If provided, and the project does not exist, it will be created. If it
    does exist, it will be set.
    :param is_public: Boolean value that sets the project's visibility. Default True.
    :param overwrite_local: If True, the current project/run log directory will be
    cleared during this function. If logging over many sessions with checkpoints, you
    may want to set this to False. Default True
    """
    if not api_client.valid_current_user():
        login()
    _init = _Init()
    BaseGalileoLogger.validate_task(task_type)
    task_type = TaskType[task_type]
    config.task_type = task_type
    _init.validate_name(project_name)
    _init.validate_name(run_name)
    if not project_name and not run_name:
        # no project and no run id, start a new project and start a new run
        project_name, run_name = random_name(), random_name()
        project_response = _init._initialize_new_project(
            project_name=project_name, is_public=is_public
        )
        run_response = _init._initialize_run_for_project(
            project_name=project_name, run_name=run_name, task_type=task_type
        )
        config.current_project_id = project_response["id"]
        config.current_run_id = run_response["id"]
        print(f"🛰 Created project, {project_name}, and new run, {run_name}.")
    elif project_name and not run_name:
        project = _init.get_project_by_name_for_user(project_name)
        # if project exists, start new run
        if project.get("id") is not None:
            run_name = random_name()
            print(f"📡 Retrieved project, {project_name}, and starting a new run")
            run_response = _init._initialize_run_for_project(
                project_name=project_name, run_name=run_name, task_type=task_type
            )
            config.current_project_id = project["id"]
            config.current_run_id = run_response["id"]
            print(
                f"🛰 Connected to project, {project_name}, and created run, {run_name}."
            )
        # otherwise create project with given name and start new run
        else:
            print(f"💭 Project {project_name} was not found.")
            run_name = random_name()
            project_response = _init._initialize_new_project(
                project_name=project_name, is_public=is_public
            )
            run_response = _init._initialize_run_for_project(
                project_name=project_name, run_name=run_name, task_type=task_type
            )
            config.current_project_id = project_response["id"]
            config.current_run_id = run_response["id"]
    elif project_name and run_name:
        project = _init.get_project_by_name_for_user(project_name)
        # if project actually exists, get the run
        if project.get("name"):
            # If the project and run exist, connect to them
            print(f"📡 Retrieving run from existing project, {project_name}")
            run = _init.get_project_run_by_name_for_user(
                project["name"], run_name=run_name
            )
            if run.get("id"):
                config.current_project_id = project["id"]
                config.current_run_id = run["id"]
                warnings.warn(
                    f"Run: {project_name}/{run_name} already exists! "
                    "The existing run will get overwritten on call to finish()!"
                )
                print(f"🛰 Connected to project, {project_name}, and run, {run_name}.")
            else:
                # If the run does not exist, create it
                run_response = _init._initialize_run_for_project(
                    project_name, run_name, task_type
                )
                config.current_project_id = project["id"]
                config.current_run_id = run_response["id"]
                print(
                    f"🛰 Connected to project, {project['name']} "
                    f"and created new run, {run_name}."
                )
        else:
            # User gave us a new project name and new run name to create, so create it
            print(f"💭 Project {project_name} was not found.")
            project_response = _init._initialize_new_project(
                project_name=project_name, is_public=is_public
            )
            run_response = _init._initialize_run_for_project(
                project_name=project_name, run_name=run_name, task_type=task_type
            )
            config.current_project_id = project_response["id"]
            config.current_run_id = run_response["id"]
            print(f"🛰 Created project, {project_name}, and new run, {run_name}.")
    else:
        # The user provided a run name and no project name. No good
        warnings.warn(
            "⚠️ You must specify a project name to initialize or create a new Galileo "
            "run. Add a project name, or simply run dataquality.init()."
        )
        return
    config.update_file_config()
    if config.current_project_id and config.current_run_id:
        _init.create_log_file_dir(
            config.current_project_id,
            config.current_run_id,
            overwrite_local=overwrite_local,
        )
    # Reset all config variables
    dataquality.get_data_logger().logger_config.reset()



"""Use dataquality client as if we were training a model without training a model
Usage: `python model_training_run.py`
To change datasets change DATASET flag to something in this s3 bucket:
https://s3.console.aws.amazon.com/s3/buckets/galileo-ml-train?region=us-west-
1&prefix=datasets/original/&showversions=false
"""

# Set environment flags for dataquality import
import os
import time
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd

os.environ["GALILEO_API_URL"] = "https://api.dev.rungalileo.io"
os.environ["GALILEO_AUTH_METHOD"] = "email"
os.environ["GALILEO_USERNAME"] = "adminy_guy@rungalileo.io"
os.environ["GALILEO_PASSWORD"] = "Admin123@"

import dataquality  # noqa
from dataquality.utils import tqdm  # noqa

DATASET = "amazon_polarity"
TRAIN_DATASET_NAME = f"{DATASET}_train.csv"
TEST_DATASET_NAME = f"{DATASET}_test.csv"
DATASET_FOLDER_PATH = Path("galileo-ml-train") / "datasets" / "original" / DATASET
TASK_TYPE = "text_classification"

NUM_EPOCHS = 1
BATCH_SIZE = 32
EMB_DIM = 768


def download_dataset_from_aws(dataset_folder_path: Union[Path, str]) -> None:
    cmd = f"aws s3 cp --recursive s3://{dataset_folder_path} ."
    os.system(cmd)


def load_dataset_split(dataset: str, split: str) -> pd.DataFrame:
    dataset = pd.read_csv(dataset + f"_{split}.csv")
    print(dataset.info(memory_usage="deep"))  # type: ignore
    return dataset


def generate_random_embeddings(batch_size: int, emb_dims: int) -> np.ndarray:
    return np.random.rand(batch_size, emb_dims)


def generate_random_probabilities(batch_size: int, num_classes: int) -> np.ndarray:
    probs = np.random.rand(batch_size, num_classes)
    return probs / probs.sum(axis=-1).reshape(-1, 1)  # Normalize to sum to 1


if __name__ == "__main__":
    download_dataset_from_aws(DATASET_FOLDER_PATH)
    train_dataset = load_dataset_split(DATASET, "train")
    test_dataset = load_dataset_split(DATASET, "test")

    dataquality.login()
    dataquality.init(
        project_name="test_large_dataset", run_name=DATASET, task_type=TASK_TYPE
    )

    t_start = time.time()
    dataquality.log_data_samples(
        texts=train_dataset["text"],
        labels=train_dataset["label"],
        ids=train_dataset["id"],
        split="train",
    )
    dataquality.log_data_samples(
        texts=test_dataset["text"],
        labels=test_dataset["label"],
        ids=test_dataset["id"],
        split="test",
    )
    dataquality.set_labels_for_run(train_dataset["label"].unique())
    print(f"Input logging took {time.time() - t_start} seconds")

    t_start = time.time()
    num_classes = train_dataset["label"].nunique()
    # Simulates model training loop
    for epoch_idx in range(NUM_EPOCHS):
        print(f"Epoch {epoch_idx}")
        # Train
        print("Training")
        for i in tqdm(range(0, len(train_dataset), BATCH_SIZE)):
            batch = train_dataset[i : i + BATCH_SIZE]

            embedding = generate_random_embeddings(len(batch), EMB_DIM)
            probs = generate_random_probabilities(len(batch), num_classes)

            dataquality.log_model_outputs(
                emb=embedding,
                probs=probs,
                split="train",
                epoch=epoch_idx,
                ids=batch["id"],
            )
        # Test
        print("Testing")
        for i in tqdm(range(0, len(test_dataset), BATCH_SIZE)):
            batch = train_dataset[i : i + BATCH_SIZE]

            embedding = generate_random_embeddings(len(batch), EMB_DIM)
            probs = generate_random_probabilities(len(batch), num_classes)

            dataquality.log_model_outputs(
                emb=embedding,
                probs=probs,
                split="test",
                epoch=epoch_idx,
                ids=batch["id"],
            )
    print(f"Took {time.time() - t_start} seconds")



import os
from threading import Thread
from time import sleep
from typing import Any, Callable, Iterable, List

from dataquality.exceptions import GalileoException


class ThreadPoolManager:
    """
    A class for managing the threaded logging calls throughout dataquality
    """

    THREADS: List[Thread] = []
    MAX_THREADS = os.cpu_count() or 10

    @staticmethod
    def add_thread(target: Callable, args: Iterable[Any] = None) -> None:
        """
        Start a new function in a thread and store that in the global list of threads

        :param target: The callable
        :param args: The arguments to the function
        :return: None
        """
        ThreadPoolManager.wait_for_thread()
        thread = Thread(target=target, args=args or [])
        try:
            thread.start()
            ThreadPoolManager.THREADS.append(thread)
        # Push up to the user
        except Exception as e:
            raise GalileoException(e)

    @staticmethod
    def wait_for_threads() -> None:
        """
        Joins all currently active threads and waits for all to be done

        :return: None
        """
        ThreadPoolManager._cleanup()
        # Waits for each thread to finish
        for i in ThreadPoolManager.THREADS:
            i.join()
        ThreadPoolManager._cleanup()

    @staticmethod
    def wait_for_thread() -> None:
        """
        Waits for an open slot in the ThreadPool for another thread.
        """
        ThreadPoolManager._cleanup()
        while len(ThreadPoolManager.THREADS) >= ThreadPoolManager.MAX_THREADS:
            sleep(0.05)
            ThreadPoolManager._cleanup()

    @staticmethod
    def _cleanup() -> None:
        """
        Cleans up the ThreadPoolManager, removing any dead ones

        :return: None
        """
        ThreadPoolManager.THREADS = [
            i for i in ThreadPoolManager.THREADS if i.is_alive()
        ]



from collections import defaultdict
from typing import Any, Dict, List, Optional, Union

import numpy as np
from scipy.special import softmax

from dataquality.loggers.logger_config.text_multi_label import (
    TextMultiLabelLoggerConfig,
    text_multi_label_logger_config,
)
from dataquality.loggers.model_logger.text_classification import (
    TextClassificationModelLogger,
)
from dataquality.schemas import __data_schema_version__


class TextMultiLabelModelLogger(TextClassificationModelLogger):
    """
    Class for logging model outputs of Multi Label Text classification models to Galileo

    * embs: (Embeddings) List[List[Union[int,float]]]. Embeddings per text sample input.
    Only one embedding vector is allowed per input (len(embs) == len(text)
    and embs.shape==2)
    * logits: Output from forward pass during model training/evalutation.
    List[List[List[float]]] or List[np.ndarray].
    For each text input, a list of lists of floats is expected (one list/array per task)
    The number of inner lists must be the number of tasks (matching the labels logged).
    The order of the inner lists is assumed to match the order of the inner list of
    labels when logging input data (matching the tasks provided by the call to
    dataquality.set_tasks_for_run()).
    * probs: (Probabilities) deprecated, use logits instead.
    * ids: Indexes of each input field: List[int]. These IDs must align with the input
    IDs for each sample input. This will be used to join them together for analysis
    by Galileo.

    ex:
    .. code-block:: python

        dq.set_epoch(0)
        dq.set_split("train")

        # 3 samples, embedding dim 768. Only 1 embedding vector can be logged for all
        # tasks. Each task CANNOT have it's own embedding vector
        embs: np.ndarray = np.random.rand(3, 768)
        # Logits per task. In this example, tasks "task_0" and "task_2" have 3 classes
        # but task "task_1" has 2
        logits: List[np.ndarray] = [
            np.random.rand(3, 3),  # 3 samples, 3 classes
            np.random.rand(3, 3),  # 3 samples, 2 classes
            np.random.rand(3, 3)  # 3 samples, 3 classes
        ]
        ids: List[int] = [0, 1, 2]

        dq.log_model_outputs(embs=embs, logits=logits, ids=ids)
    """

    __logger_name__ = "text_multi_label"
    logger_config: TextMultiLabelLoggerConfig = (
        text_multi_label_logger_config  # type: ignore
    )

    def __init__(
        self,
        embs: Union[List, np.ndarray] = None,
        probs: Union[List[List[List]], List[np.ndarray]] = None,
        logits: Union[List[List[List]], List[np.ndarray]] = None,
        ids: Union[List, np.ndarray] = None,
        split: str = "",
        epoch: Optional[int] = None,
        inference_name: Optional[str] = None,
    ) -> None:
        super().__init__(
            embs=embs,
            probs=probs,
            logits=logits,
            ids=ids,
            split=split,
            epoch=epoch,
            inference_name=inference_name,
        )

    def validate(self) -> None:
        """
        Validates that the current config is correct.
        * embs, probs, and ids must exist and be the same length
        :return:
        """
        super().validate()
        for ind, probs_per_task in enumerate(self.probs):
            assert len(probs_per_task) == self.logger_config.observed_num_tasks, (
                f"Expected {self.logger_config.observed_num_tasks} probability vectors "
                f"per input (based on input data logging) but found "
                f"{len(probs_per_task)} for input {ind}."
            )
            for task_ind, task_probs in enumerate(probs_per_task):
                assert (
                    len(task_probs) > 0
                ), f"Cannot log empty probability list for task {task_ind}."
        # For each record, the task probability vector should be the same length
        for task_ind in range(self.logger_config.observed_num_tasks):
            num_prob_per_task = set([len(p[task_ind]) for p in self.probs])
            assert len(num_prob_per_task) == 1, (
                f"Task {task_ind} has an inconsistent number of probabilities in this "
                f"batch. Found {num_prob_per_task} unique number of labels. "
                f"Every input for a given task should have the same number of "
                f"values in its probability vector."
            )

    def _get_data_dict(self) -> Dict[str, Any]:
        data = defaultdict(list)
        for record_id, prob_per_task, emb in zip(self.ids, self.probs, self.embs):
            # Handle binary classification by making it 2-class classification
            record = {
                "id": record_id,
                "epoch": self.epoch,
                "split": self.split,
                "emb": emb,
                "data_schema_version": __data_schema_version__,
            }
            # Break out the probabilities and predictions into a col per task
            for task_num in range(self.logger_config.observed_num_tasks):
                task_probs: List[float] = prob_per_task[task_num]
                if len(task_probs) == 1:  # Handle binary classification case
                    task_probs = [1 - task_probs[0], task_probs[0]]
                record[f"prob_{task_num}"] = np.array(task_probs, dtype=np.float32)
                record[f"pred_{task_num}"] = int(np.argmax(task_probs))

            for k in record.keys():
                data[k].append(record[k])
        return data

    def _set_num_labels(self, data: Dict) -> None:
        num_labels_per_task = []
        for task_num in range(self.logger_config.observed_num_tasks):
            len(data[f"prob_{task_num}"][0])
            num_labels = len(data[f"prob_{task_num}"][0])
            num_labels_per_task.append(num_labels)
        self.logger_config.observed_num_labels = num_labels_per_task

    def __setattr__(self, key: Any, value: Any) -> None:
        if key not in self.get_valid_attributes() and not key.startswith("prob_"):
            raise AttributeError(
                f"{key} is not a valid attribute of {self.__logger_name__} logger. "
                f"Only {self.get_valid_attributes()}"
            )
        super().__setattr__(key, value)

    def convert_logits_to_prob_binary(self, sample_logits: np.ndarray) -> np.ndarray:
        """Converts logits to probs in the binary case

        Takes the sigmoid of the single class logits and adds the negative
        lass prediction (1-class pred)
        """
        assert sample_logits.ndim == 2, (
            f"In binary multi-label, your logits should have 2 dimensions, but "
            f"they currently have {sample_logits.ndim}. Do you mean to use to "
            f"binary multi-label? If not, call dq.set_tasks_for_run(tasks) without "
            f"the binary=True flag. Or call dq.init() to reset."
        )
        return super().convert_logits_to_prob_binary(sample_logits)

    def convert_logits_to_probs(
        self, sample_logits: Union[List, np.ndarray]
    ) -> np.ndarray:
        """Converts logits to probs via softmax per sample

        In the case of binary multi-label, we don't run softmax, we use sigmoid
        """
        if self.logger_config.binary:
            return self.convert_logits_to_prob_binary(
                np.array(sample_logits, dtype=np.float32)
            )
        # axis ensures that in a matrix of probs with dims num_samples x num_classes
        # we take the softmax for each sample
        probs = []
        for sample_logits in sample_logits:
            task_probs = []
            for task_logits in sample_logits:
                task_logits = self._convert_tensor_ndarray(task_logits)
                task_probs.append(softmax(task_logits.astype(np.float_), axis=-1))
            probs.append(task_probs)
        return np.array(probs, dtype=object)



try:
    from IPython import get_ipython

    if get_ipython():
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
except ImportError:
    from tqdm import tqdm

tqdm = tqdm

__all__ = ["tqdm"]



import getpass
import os
from typing import Callable, Dict

import requests

from dataquality.clients.api import ApiClient
from dataquality.core._config import AuthMethod, Config, config
from dataquality.exceptions import GalileoException
from dataquality.utils.helpers import check_noop

GALILEO_AUTH_METHOD = "GALILEO_AUTH_METHOD"
api_client = ApiClient()


class _Auth:
    def auth_methods(self) -> Dict[AuthMethod, Callable]:
        return {AuthMethod.email: self.email_login}

    def email_login(self) -> None:
        if self.email_token_present_and_valid(config):
            return

        username = os.getenv("GALILEO_USERNAME")
        password = os.getenv("GALILEO_PASSWORD")
        if not username or not password:
            username = input("📧 Enter your email:")
            password = getpass.getpass("🤫 Enter your password:")
        res = requests.post(
            f"{config.api_url}/login",
            data={
                "username": username,
                "password": password,
                "auth_method": config.auth_method,
            },
            headers={"X-Galileo-Request-Source": "dataquality_python_client"},
        )
        if res.status_code != 200:
            raise GalileoException(f"Issue authenticating: {res.json()['detail']}")

        access_token = res.json().get("access_token")
        config.token = access_token
        config.update_file_config()

    def email_token_present_and_valid(self, config: Config) -> bool:
        return config.auth_method == "email" and api_client.valid_current_user()


@check_noop
def login() -> None:
    print(f"📡 {config.api_url.replace('api.','console.')}")
    print("🔭 Logging you into Galileo\n")
    auth_methods = ",".join([am.value for am in AuthMethod])
    # Try auto auth config
    if len(list(AuthMethod)) == 1:
        auth_method = list(AuthMethod)[0].value
        os.environ[GALILEO_AUTH_METHOD] = auth_method
    auth_method = os.getenv(GALILEO_AUTH_METHOD) or config.auth_method
    if not auth_method or auth_method.lower() not in list(AuthMethod):
        # We currently only have 1 auth method, so no reason to ask the user
        if len(list(AuthMethod)) == 1:
            auth_method = list(AuthMethod)[0].value
        else:
            auth_method = input(
                "🔐 How would you like to login? \n"
                f"Enter one of the following: {auth_methods}\n"
            )
        if auth_method.lower() not in list(AuthMethod):
            print(
                "Invalid login request. You must input one of "
                f"the following authentication methods: {auth_methods}."
            )
            return
        else:  # Save it as an environment variable for the next login
            print("🤝 Saving preferred login method")
            os.environ[GALILEO_AUTH_METHOD] = auth_method
    else:
        print(f"👀 Found auth method {auth_method} set via env, skipping prompt.")
    config.auth_method = AuthMethod(auth_method)
    _auth = _Auth()
    _auth.auth_methods()[config.auth_method]()
    current_user_email = api_client.get_current_user().get("email")
    if not current_user_email:
        return
    config.current_user = current_user_email
    config.update_file_config()
    print(f"🚀 You're logged in to Galileo as {current_user_email}!")



from dataquality.loggers.model_logger import (
    text_classification,
    text_multi_label,
    text_ner,
)
from dataquality.loggers.model_logger.base_model_logger import BaseGalileoModelLogger

__all__ = [
    "text_classification",
    "text_multi_label",
    "text_ner",
    "BaseGalileoModelLogger",
]



"dataquality"

__version__ = "v0.3.8"

import os
import resource

import dataquality.core._config
import dataquality.metrics
from dataquality.core._config import config
from dataquality.core.auth import login
from dataquality.core.finish import finish, get_run_status, wait_for_run
from dataquality.core.init import init
from dataquality.core.log import (
    docs,
    get_data_logger,
    get_model_logger,
    log_data_sample,
    log_data_samples,
    log_dataset,
    log_model_outputs,
    set_epoch,
    set_labels_for_run,
    set_split,
    set_tagging_schema,
    set_tasks_for_run,
)
from dataquality.utils.dq_logger import get_dq_log_file
from dataquality.utils.helpers import check_noop


@check_noop
def configure() -> None:
    """Update your active config with new env variables.
    Reset user token on configure and prompt new login.

    Available environment variables to update:
    * GALILEO_CONSOLE_URL
    """
    if "GALILEO_API_URL" in os.environ:
        del os.environ["GALILEO_API_URL"]
    updated_config = dataquality.core._config.reset_config()
    for k, v in updated_config.dict().items():
        config.__setattr__(k, v)
    config.token = None
    config.update_file_config()
    login()


__all__ = [
    "__version__",
    "login",
    "init",
    "log_data_samples",
    "log_model_outputs",
    "config",
    "configure",
    "finish",
    "set_labels_for_run",
    "get_data_logger",
    "get_model_logger",
    "set_tasks_for_run",
    "set_tagging_schema",
    "docs",
    "wait_for_run",
    "get_run_status",
    "set_epoch",
    "set_split",
    "log_data_sample",
    "log_dataset",
    "get_dq_log_file",
]

try:
    resource.setrlimit(resource.RLIMIT_NOFILE, (65535, 65535))
except ValueError:  # The users limit is higher than our max, which is OK
    pass



from typing import Callable

import pytest
import pytorch_lightning as pl
import torch

import dataquality
from dataquality.integrations.lightning import DataQualityCallback
from dataquality.integrations.torch import log_input_data, watch
from dataquality.schemas.split import Split
from dataquality.utils.thread_pool import ThreadPoolManager
from tests.test_dataquality import validate_uploaded_data
from tests.utils.data_utils import validate_cleanup_data
from tests.utils.lightning_model import (
    NUM_RECORDS,
    NewsgroupDataset,
    model,
    torch_model,
)


@pytest.mark.noautofixt
def test_lightning_autolog(
    cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """
    Tests the lightning autolog config and that data is properly stored / logged
    """
    train_dataloader = torch.utils.data.DataLoader(
        NewsgroupDataset("training"), batch_size=NUM_RECORDS, shuffle=True
    )
    test_dataloader = torch.utils.data.DataLoader(
        NewsgroupDataset("test"), batch_size=NUM_RECORDS, shuffle=True
    )
    trainer = pl.Trainer(
        max_epochs=1, num_sanity_val_steps=0, callbacks=[(DataQualityCallback())]
    )

    trainer.fit(model, train_dataloader)  # Will get logged as epoch 0
    trainer.test(model, test_dataloader)  # Will get logged as epoch 0
    ThreadPoolManager.wait_for_threads()
    # Mock call to finish
    logger = dataquality.get_data_logger()
    logger.upload()
    validate_uploaded_data(expected_num_records=NUM_RECORDS)
    logger._cleanup()
    validate_cleanup_data()


@pytest.mark.noautofixt
def test_torch_autolog(cleanup_after_use: Callable, set_test_config: Callable) -> None:
    """Tests our watch(model) functionality for pytorch"""
    train_dataloader = torch.utils.data.DataLoader(
        NewsgroupDataset("training"), batch_size=NUM_RECORDS, shuffle=True
    )
    test_dataloader = torch.utils.data.DataLoader(
        NewsgroupDataset("test"), batch_size=NUM_RECORDS, shuffle=True
    )
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    torch_model.to(device)

    log_input_data(train_dataloader, "training")
    log_input_data(test_dataloader, "test")

    watch(torch_model)

    for epoch in range(2):
        dataquality.set_epoch(epoch)
        dataquality.set_split(Split.training)
        torch_model.train()
        for idx, data in enumerate(train_dataloader):
            x_idxs, x, attention_mask, y = data
            x = x.to(device)
            attention_mask = attention_mask.to(device)
            y = torch.tensor(y, device=device)

            # forward + backward + optimize
            _ = torch_model(x, attention_mask, x_idxs=x_idxs)

        with torch.no_grad():
            dataquality.set_split(Split.testing)
            for data in test_dataloader:
                x_idxs, x, attention_mask, y = data

                x = x.to(device)
                attention_mask = attention_mask.to(device)
                _ = torch_model(x, attention_mask, x_idxs=x_idxs)

    ThreadPoolManager.wait_for_threads()
    # Mock call to finish
    logger = dataquality.get_data_logger()
    logger.upload()
    validate_uploaded_data(expected_num_records=NUM_RECORDS)
    logger._cleanup()
    validate_cleanup_data()



from dataquality.utils.name import ADJECTIVES, ANIMALS, COLORS, random_name


def test_name() -> None:
    adj, c, anm = random_name().split("_")
    assert adj in ADJECTIVES
    assert c in COLORS
    assert anm in ANIMALS



from typing import Dict, List, Union
from uuid import uuid4

import dataquality
from dataquality import __version__

config = dataquality.config

EXISTING_PROJECT = "existing_proj"
EXISTING_RUN = "existing_run"
TMP_CREATE_NEW_PROJ_RUN = None
FAKE_NEW_RUN = uuid4()


class MockResponse:
    def __init__(
        self, json_data: Union[Dict, List], status_code: int, headers: Dict = None
    ) -> None:
        self.headers = headers or {}
        self.json_data = json_data
        self.status_code = status_code
        if status_code in (200, 204):
            self.ok = True
        else:
            self.ok = False
            self.text = json_data

    def json(self) -> Union[Dict, List]:
        return self.json_data

    def raise_for_status(self) -> None:
        pass


def mocked_healthcheck_request(request_url: str) -> MockResponse:
    if request_url.endswith("healthcheck/runners"):
        return MockResponse({"api_version": __version__}, 200)

    return MockResponse({}, 200)


def mocked_healthcheck_request_new_api_version(request_url: str) -> MockResponse:
    if request_url.endswith("healthcheck/runners"):
        return MockResponse({"api_version": "100.0.0"}, 200)
    return MockResponse({}, 200)


def mocked_login_requests(
    request_url: str,
    json: Dict = {},
    params: Dict = {},
    headers: Dict = {},
    data: Dict = {},
) -> MockResponse:
    if request_url.endswith("login"):
        return MockResponse(
            {
                "access_token": "mock_token",
            },
            200,
        )
    if request_url.endswith("current_user"):
        return MockResponse({"user": "user"}, 200)
    return MockResponse({}, 200)


def mocked_failed_login_requests(
    request_url: str,
    json: Dict = {},
    params: Dict = {},
    headers: Dict = {},
    data: Dict = {},
) -> MockResponse:
    if request_url.endswith("login"):
        return MockResponse({"detail": "Incorrect login credentials."}, 404)
    return MockResponse({}, 404)


def mocked_get_project_run(
    request_url: str, json: Dict, params: Dict, headers: Dict, data: Dict
) -> MockResponse:
    if request_url.endswith("current_user"):
        return MockResponse({"id": "user"}, 200)
    res = [
        {"id": uuid4(), "name": EXISTING_PROJECT, "project_id": uuid4()},
        {"id": uuid4(), "name": EXISTING_RUN, "project_id": uuid4()},
        {"id": uuid4(), "name": TMP_CREATE_NEW_PROJ_RUN, "project_id": uuid4()},
    ]
    return MockResponse(res, 200)


def mocked_create_project_run(
    request_url: str, json: Dict, params: Dict, headers: Dict, data: Dict
) -> MockResponse:
    global TMP_CREATE_NEW_PROJ_RUN
    TMP_CREATE_NEW_PROJ_RUN = json["name"]
    res = {"id": uuid4(), "name": TMP_CREATE_NEW_PROJ_RUN}
    return MockResponse(res, 200)


def mocked_missing_run(
    request_url: str, json: Dict, params: Dict, headers: Dict, data: Dict
) -> MockResponse:
    if request_url.endswith("current_user"):
        return MockResponse({"id": "user"}, 200)
    # Run does not exist
    if "run_name" in request_url:
        return MockResponse([], 204)
    # Project does exist
    else:
        res = {"id": uuid4(), "name": EXISTING_PROJECT}
        return MockResponse([res], 200)


def mocked_missing_project_run(
    request_url: str, json: Dict, params: Dict, headers: Dict, data: Dict
) -> MockResponse:
    if request_url.endswith("current_user"):
        return MockResponse({"id": "user"}, 200)
    return MockResponse([{"id": uuid4(), "name": TMP_CREATE_NEW_PROJ_RUN}], 200)


def mocked_missing_project_name(
    request_url: str, json: Dict, params: Dict, headers: Dict, data: Dict
) -> MockResponse:
    if request_url.endswith("current_user"):
        return MockResponse({"id": "user"}, 200)
    return MockResponse([], 200)


def mocked_delete_project_run(
    request_url: str, json: Dict, params: Dict, headers: Dict, data: Dict
) -> MockResponse:
    if request_url.endswith("current_user"):
        return MockResponse({"id": "user"}, 200)
    return MockResponse([{"id": uuid4(), "name": TMP_CREATE_NEW_PROJ_RUN}], 200)


def mocked_delete_project_not_found(
    request_url: str, json: Dict, params: Dict, headers: Dict, data: Dict
) -> MockResponse:
    if request_url.endswith("current_user"):
        return MockResponse({"id": "user"}, 200)
    return MockResponse({"detail": "project not found"}, 404)


def mocked_login() -> None:
    config.token = "mock_token"
